from __future__ import annotations
import enum
from typing import Any
from bcsfe import core


class MaxValueType(enum.Enum):
    CATFOOD = "catfood"
    XP = "xp"
    NORMAL_TICKETS = "normal_tickets"
    HUNDRED_MILLION_TICKETS = "100_million_tickets"
    RARE_TICKETS = "rare_tickets"
    PLATINUM_TICKETS = "platinum_tickets"
    LEGEND_TICKETS = "legend_tickets"
    NP = "np"
    LEADERSHIP = "leadership"
    BATTLE_ITEMS = "battle_items"
    CATAMINS = "catamins"
    CATSEYES = "catseyes"
    CATFRUIT = "catfruit"
    BASE_MATERIALS = "base_materials"
    LABYRINTH_MEDALS = "labyrinth_medals"
    TALENT_ORBS = "talent_orbs"
    TREASURE_LEVEL = "treasure_level"
    STAGE_CLEAR_COUNT = "stage_clear_count"
    ITF_TIMED_SCORE = "itf_timed_score"
    EVENT_TICKETS = "event_tickets"
    TREASURE_CHESTS = "treasure_chests"


class MaxValueHelper:
    def __init__(self):
        self.max_value_data = self.get_max_value_data()

    @staticmethod
    def convert_val_code(value_code: MaxValueType | str) -> str:
        if isinstance(value_code, MaxValueType):
            value_code = value_code.value
        return value_code

    def get_max_value_data(self) -> dict[str, Any]:
        file_path = core.Path("max_values.json", True)
        if not file_path.exists():
            return {}
        try:
            return core.JsonFile.from_data(file_path.read()).to_object()
        except core.JSONDecodeError:
            return {}

    def get(self, value_code: str | MaxValueType) -> int:
        try:
            return int(self.max_value_data.get(self.convert_val_code(value_code), 0))
        except ValueError:
            return 0

    def get_property(self, value_code: str | MaxValueType, property: str) -> int:
        try:
            return int(
                self.max_value_data.get(self.convert_val_code(value_code), {}).get(
                    property, 0
                )
            )
        except ValueError:
            return 0

    def get_old(self, value_code: str | MaxValueType) -> int:
        return self.get_property(value_code, "old")

    def get_new(self, value_code: str | MaxValueType) -> int:
        return self.get_property(value_code, "new")



from __future__ import annotations
from typing import Any

from bcsfe import core
import datetime
import json
from bcsfe.cli import color, dialog_creator


class RecentSave:
    def __init__(
        self,
        path: core.Path,
        cc: core.CountryCode,
        gv: core.GameVersion,
        inquiry: str,
        time: datetime.datetime,
        name: core.Path,
    ):
        self.path = path
        self.cc = cc
        self.gv = gv
        self.inquiry = inquiry
        self.time = time
        self.name = name

    @staticmethod
    def from_dict(data: dict[str, Any]) -> RecentSave | None:
        path = data.get("path")
        cc = data.get("cc")
        gv = data.get("gv")
        inquiry = data.get("inquiry")
        time_stamp = data.get("timestamp")
        name = data.get("name")
        if (
            path is None
            or cc is None
            or gv is None
            or inquiry is None
            or time_stamp is None
            or name is None
        ):
            return None

        return RecentSave(
            core.Path(path),
            core.CountryCode(cc),
            core.GameVersion.from_string(gv),
            inquiry,
            datetime.datetime.fromtimestamp(time_stamp),
            core.Path(name),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "path": self.path.to_str(),
            "cc": self.cc.get_code(),
            "gv": self.gv.to_string(),
            "inquiry": self.inquiry,
            "timestamp": self.time.timestamp(),
            "name": self.name.to_str(),
        }


class RecentSaves:
    def __init__(self, saves: list[RecentSave]):
        self.saves = saves

    @staticmethod
    def from_json(data: list[dict[str, Any]]) -> RecentSaves:
        res: list[RecentSave] = []

        for item in data:
            save = RecentSave.from_dict(item)
            if save is not None:
                res.append(save)

        return RecentSaves(res)

    def to_json(self) -> list[dict[str, Any]]:
        return [save.to_dict() for save in self.saves][-10:]  # only store 10

    @staticmethod
    def from_path(path: core.Path) -> RecentSaves | None:
        json_data = json.loads(path.read().to_str())

        return RecentSaves.from_json(json_data)

    def to_path(self, path: core.Path):
        data = json.dumps(self.to_json(), indent=4)

        path.write(core.Data(data))

    @staticmethod
    def read_default() -> RecentSaves:
        path = RecentSaves.get_path()
        if path.exists():
            return RecentSaves.from_path(path) or RecentSaves([])
        return RecentSaves([])

    @staticmethod
    def get_path() -> core.Path:
        return core.Path.get_documents_folder().add("recent_saves.json")

    def save_default(self):
        path = RecentSaves.get_path()
        self.to_path(path)

    def select(self) -> RecentSave | None:
        if not self.saves:
            color.ColoredText.localize("no_recent_saves")
            return None
        items: list[str] = []
        for save in self.saves:
            items.append(
                core.localize(
                    "recent_save",
                    path=save.path,
                    cc=save.cc,
                    gv=save.gv,
                    inquiry_code=save.inquiry,
                    year=save.time.year,
                    month=str(save.time.month).zfill(2),
                    day=str(save.time.day).zfill(2),
                    hour=str(save.time.hour).zfill(2),
                    minute=str(save.time.minute).zfill(2),
                    second=str(save.time.second).zfill(2),
                    name=save.name,
                )
            )
        items.reverse()

        resp = dialog_creator.ChoiceInput.from_reduced(
            items, localize_options=False, dialog="select_recent"
        ).single_choice()
        if resp is None:
            return None

        resp = len(self.saves) - resp

        return self.saves[resp]



from __future__ import annotations
from bcsfe import core
from bcsfe.cli import color, dialog_creator
from bcsfe.cli.edits import cat_editor


def display_storage(save_file: core.SaveFile, storage: list[core.StorageItem]):
    color.ColoredText.localize("current_storage_items")
    index = 0
    for item in storage:
        if item.item_type == 0:
            continue

        index += 1
        color.ColoredText(f"{index}. ", end="")
        display_item(item, save_file)

    if index == 0:
        color.ColoredText.localize("storage_is_empty")

    available_slots = len(storage) - index

    color.ColoredText.localize("available_storage", slots=available_slots)


def display_item(item: core.StorageItem, save_file: core.SaveFile):
    color.ColoredText(get_item_str(item, save_file))


def get_item_str(item: core.StorageItem, save_file: core.SaveFile) -> str:
    if item.item_type == 1:
        cat_id = item.item_id
        names = core.Cat.get_names(cat_id, save_file)

        if not names:
            names = [str(cat_id)]

        return core.localize("cat", name=names[0], id=cat_id)
    elif item.item_type == 2:
        skill_id = item.item_id

        skill_names = (
            core.core_data.get_gatya_item_buy(save_file).get_names_by_category(
                core.GatyaItemCategory.SPECIAL_SKILLS
            )
            or []
        )

        if skill_id >= len(skill_names) or skill_id < 0:
            name = str(skill_id)
        else:
            name = skill_names[skill_id][1]

        return core.localize("special_skill", name=name, id=skill_id)
    elif item.item_type == 3:
        item_id = item.item_id

        name = core.core_data.get_gatya_item_names(save_file).get_name(item_id)
        if name is None:
            name = str(item_id)

        return core.localize("item", name=name, id=item_id)
    else:
        return core.localize(
            "unrecognised_storage_item", item_type=item.item_type, id=item.item_id
        )


def clear_storage(storage: list[core.StorageItem]):
    for item in storage:
        item.item_id = 0
        item.item_type = 0


def add_item(storage: list[core.StorageItem], item: core.StorageItem) -> bool:
    for citem in storage:
        if citem.item_type == 0:
            citem.item_type = item.item_type
            citem.item_id = item.item_id
            return True
    return False


def get_storage_space(storage: list[core.StorageItem]) -> int:
    space = 0

    for item in storage:
        if item.item_type == 0:
            space += 1
    return space


def edit_storage(save_file: core.SaveFile):
    display_storage(save_file, save_file.cats.storage_items)
    exit = False
    while not exit:
        exit = edit_loop(save_file)

    color.ColoredText.localize("storage_success")


def edit_loop(save_file: core.SaveFile) -> bool:
    storage = save_file.cats.storage_items

    options = [
        "display_storage",
        "clear_storage",
        "add_cats",
        "add_special_skills",
        "remove_items",
        "finish",
    ]

    choice = dialog_creator.ChoiceInput.from_reduced(
        options, dialog="select_option"
    ).single_choice()
    if choice is None:
        return False

    choice -= 1

    if choice == 0:
        display_storage(save_file, storage)
    if choice == 1:
        clear_storage(storage)
    elif choice == 2:
        editor, cats = cat_editor.CatEditor.from_save_file(save_file)
        if editor is None:
            return False

        space = get_storage_space(storage)
        if len(cats) > len(storage):
            color.ColoredText.localize(
                "too_many_cats_selected", max=len(storage), current=len(cats)
            )
            return False

        needs = len(cats) - space
        if needs > 0:
            color.ColoredText.localize("need_x_more_space", needs=needs)
            return False

        color.ColoredText.localize("added_cats")
        for cat in cats:
            item = core.StorageItem.from_cat(cat.id)
            add_item(storage, item)
            display_item(item, save_file)
    elif choice == 3:

        skill_names: list[str] = list(
            map(
                lambda sk: sk[1] or str(sk[0].id),
                core.core_data.get_gatya_item_buy(save_file).get_names_by_category(
                    core.GatyaItemCategory.SPECIAL_SKILLS
                )
                or [],
            )
        )

        options, _ = dialog_creator.ChoiceInput.from_reduced(
            skill_names, localize_options=False, dialog="select_special_skills"
        ).multiple_choice(False)

        if options is None:
            return False

        space = get_storage_space(storage)
        if len(options) > len(storage):
            color.ColoredText.localize(
                "too_many_skills_selected", max=len(storage), current=len(options)
            )
            return False

        needs = len(options) - space
        if needs > 0:
            color.ColoredText.localize("need_x_more_space", needs=needs)
            return False

        color.ColoredText.localize("added_special_skills")
        for choice in options:
            item = core.StorageItem.from_special_skill(choice)
            add_item(storage, item)
            display_item(item, save_file)

    elif choice == 4:
        options2: list[str] = []
        for item in storage:
            if item.item_type == 0:
                continue
            options2.append(get_item_str(item, save_file))

        choices, _ = dialog_creator.ChoiceInput.from_reduced(
            options2, localize_options=False
        ).multiple_choice(False)
        if choices is None:
            return False

        color.ColoredText.localize("removed_items")
        index = 0
        for item in storage:
            if item.item_type == 0:
                continue

            if index in choices:
                display_item(item, save_file)
                item.item_type = 0
                item.item_id = 0

            index += 1

    elif choice == 5:
        return True

    return False



from __future__ import annotations
from bcsfe import core


class Stage:
    def __init__(self, score: int):
        self.score = score

    @staticmethod
    def init() -> Stage:
        return Stage(0)

    @staticmethod
    def read(stream: core.Data) -> Stage:
        return Stage(stream.read_int())

    def write(self, stream: core.Data):
        stream.write_int(self.score)

    def serialize(self) -> int:
        return self.score

    @staticmethod
    def deserialize(data: int) -> Stage:
        return Stage(data)

    def __repr__(self) -> str:
        return f"Stage(score={self.score})"

    def __str__(self) -> str:
        return self.__repr__()


class SubChapter:
    def __init__(self, stages: list[Stage]):
        self.stages = stages

    @staticmethod
    def init(total_stages: int) -> SubChapter:
        return SubChapter([Stage.init() for _ in range(total_stages)])

    @staticmethod
    def read(stream: core.Data, total_stages: int) -> SubChapter:
        stages: list[Stage] = []
        for _ in range(total_stages):
            stages.append(Stage.read(stream))
        return SubChapter(stages)

    def write(self, stream: core.Data):
        for stage in self.stages:
            stage.write(stream)

    def serialize(self) -> list[int]:
        return [stage.serialize() for stage in self.stages]

    @staticmethod
    def deserialize(data: list[int]) -> SubChapter:
        return SubChapter([Stage.deserialize(stage) for stage in data])

    def __repr__(self) -> str:
        return f"SubChapter(stages={self.stages})"

    def __str__(self) -> str:
        return self.__repr__()


class SubChapterStars:
    def __init__(self, sub_chapters: list[SubChapter]):
        self.sub_chapters = sub_chapters

    @staticmethod
    def init(total_stages: int, total_stars: int) -> SubChapterStars:
        return SubChapterStars(
            [SubChapter.init(total_stages) for _ in range(total_stars)]
        )

    @staticmethod
    def read(
        stream: core.Data,
        total_stages: int,
        total_stars: int,
    ) -> SubChapterStars:
        sub_chapters: list[SubChapter] = []
        for _ in range(total_stars):
            sub_chapters.append(SubChapter.read(stream, total_stages))
        return SubChapterStars(sub_chapters)

    def write(self, stream: core.Data):
        for sub_chapter in self.sub_chapters:
            sub_chapter.write(stream)

    def serialize(self) -> list[list[int]]:
        return [sub_chapter.serialize() for sub_chapter in self.sub_chapters]

    @staticmethod
    def deserialize(data: list[list[int]]) -> SubChapterStars:
        return SubChapterStars(
            [SubChapter.deserialize(sub_chapter) for sub_chapter in data]
        )

    def __repr__(self) -> str:
        return f"SubChapterStars(sub_chapters={self.sub_chapters})"

    def __str__(self) -> str:
        return self.__repr__()


class TimedScoreChapters:
    def __init__(self, sub_chapters: list[SubChapterStars]):
        self.sub_chapters = sub_chapters

    @staticmethod
    def init(gv: core.GameVersion) -> TimedScoreChapters:
        if gv < 20:
            return TimedScoreChapters([])
        if gv <= 33:
            total_subchapters = 50
            total_stages = 12
            total_stars = 3
        elif gv <= 34:
            total_subchapters = 0
            total_stages = 12
            total_stars = 3
        else:
            total_subchapters = 0
            total_stages = 0
            total_stars = 0
        return TimedScoreChapters(
            [
                SubChapterStars.init(total_stages, total_stars)
                for _ in range(total_subchapters)
            ]
        )

    @staticmethod
    def read(stream: core.Data, gv: core.GameVersion) -> TimedScoreChapters:
        if gv < 20:
            return TimedScoreChapters([])
        if gv <= 33:
            total_subchapters = 50
            total_stages = 12
            total_stars = 3
        elif gv <= 34:
            total_subchapters = stream.read_int()
            total_stages = 12
            total_stars = 3
        else:
            total_subchapters = stream.read_int()
            total_stages = stream.read_int()
            total_stars = stream.read_int()
        sub_chapters: list[SubChapterStars] = []
        for _ in range(total_subchapters):
            sub_chapters.append(
                SubChapterStars.read(stream, total_stages, total_stars)
            )
        return TimedScoreChapters(sub_chapters)

    def write(self, stream: core.Data, gv: core.GameVersion):
        if gv < 20:
            return
        if gv <= 33:
            pass
        elif gv <= 34:
            stream.write_int(len(self.sub_chapters))
        else:
            stream.write_int(len(self.sub_chapters))
            try:
                stream.write_int(
                    len(self.sub_chapters[0].sub_chapters[0].stages)
                )
            except IndexError:
                stream.write_int(0)
            try:
                stream.write_int(len(self.sub_chapters[0].sub_chapters))
            except IndexError:
                stream.write_int(0)
        for sub_chapter in self.sub_chapters:
            sub_chapter.write(stream)

    def serialize(self) -> list[list[list[int]]]:
        return [sub_chapter.serialize() for sub_chapter in self.sub_chapters]

    @staticmethod
    def deserialize(data: list[list[list[int]]]) -> TimedScoreChapters:
        return TimedScoreChapters(
            [SubChapterStars.deserialize(sub_chapter) for sub_chapter in data]
        )

    def __repr__(self) -> str:
        return f"Chapters(sub_chapters={self.sub_chapters})"

    def __str__(self) -> str:
        return self.__repr__()



from __future__ import annotations
from bcsfe import core
from typing import Any


class CatSlot:
    def __init__(self, cat_id: int, form: int):
        self.cat_id = cat_id
        self.form = form

    @staticmethod
    def init() -> CatSlot:
        return CatSlot(0, 0)

    @staticmethod
    def read(stream: core.Data) -> CatSlot:
        cat_id = stream.read_short()
        form = stream.read_byte()
        return CatSlot(cat_id, form)

    def write(self, stream: core.Data):
        stream.write_short(self.cat_id)
        stream.write_byte(self.form)

    def serialize(self) -> dict[str, Any]:
        return {
            "cat_id": self.cat_id,
            "form": self.form,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> CatSlot:
        return CatSlot(data.get("cat_id", 0), data.get("form", 0))

    def __repr__(self):
        return f"CatSlot({self.cat_id}, {self.form})"

    def __str__(self):
        return self.__repr__()


class LineupCat:
    def __init__(
        self,
        index: int,
        cats: list[CatSlot],
        u1: int,
        u2: int,
        u3: int,
    ):
        self.index = index
        self.cats = cats
        self.u1 = u1
        self.u2 = u2
        self.u3 = u3

    @staticmethod
    def init() -> LineupCat:
        cats = [CatSlot.init() for _ in range(10)]
        return LineupCat(0, cats, 0, 0, 0)

    @staticmethod
    def read(stream: core.Data) -> LineupCat:
        index = stream.read_short()
        length = 10

        cats = [CatSlot.read(stream) for _ in range(length)]
        u1 = stream.read_byte()
        u2 = stream.read_byte()
        u3 = stream.read_byte()
        return LineupCat(index, cats, u1, u2, u3)

    def write(self, stream: core.Data):
        stream.write_short(self.index)
        for cat in self.cats:
            cat.write(stream)
        stream.write_byte(self.u1)
        stream.write_byte(self.u2)
        stream.write_byte(self.u3)

    def serialize(self) -> dict[str, Any]:
        return {
            "index": self.index,
            "cats": [cat.serialize() for cat in self.cats],
            "u1": self.u1,
            "u2": self.u2,
            "u3": self.u3,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> LineupCat:
        return LineupCat(
            data.get("index", 0),
            [CatSlot.deserialize(cat) for cat in data.get("cats", [])],
            data.get("u1", 0),
            data.get("u2", 0),
            data.get("u3", 0),
        )

    def __repr__(self):
        return f"LineupCat({self.index}, {self.cats}, {self.u1}, {self.u2}, {self.u3})"

    def __str__(self):
        return self.__repr__()


class ClearedSlotsCat:
    def __init__(self, lineups: list[LineupCat]):
        self.lineups = lineups

    @staticmethod
    def init() -> ClearedSlotsCat:
        return ClearedSlotsCat([])

    @staticmethod
    def read(stream: core.Data) -> ClearedSlotsCat:
        total = stream.read_short()
        lineups = [LineupCat.read(stream) for _ in range(total)]
        return ClearedSlotsCat(lineups)

    def write(self, stream: core.Data):
        stream.write_short(len(self.lineups))
        for lineup in self.lineups:
            lineup.write(stream)

    def serialize(self) -> list[dict[str, Any]]:
        return [lineup.serialize() for lineup in self.lineups]

    @staticmethod
    def deserialize(data: list[dict[str, Any]]) -> ClearedSlotsCat:
        return ClearedSlotsCat(
            [LineupCat.deserialize(lineup) for lineup in data],
        )

    def __repr__(self):
        return f"ClearedSlotsCat({self.lineups})"

    def __str__(self):
        return self.__repr__()


class StageSlot:
    def __init__(self, stage_id: int):
        self.stage_id = stage_id

    @staticmethod
    def init() -> StageSlot:
        return StageSlot(0)

    @staticmethod
    def read(stream: core.Data) -> StageSlot:
        stage_id = stream.read_int()
        return StageSlot(stage_id)

    def write(self, stream: core.Data):
        stream.write_int(self.stage_id)

    def serialize(self) -> int:
        return self.stage_id

    @staticmethod
    def deserialize(data: int) -> StageSlot:
        return StageSlot(data)

    def __repr__(self):
        return f"StageSlot({self.stage_id})"

    def __str__(self):
        return self.__repr__()


class StageLineups:
    def __init__(self, index: int, slots: list[StageSlot]):
        self.index = index
        self.slots = slots

    @staticmethod
    def init() -> StageLineups:
        return StageLineups(0, [])

    @staticmethod
    def read(stream: core.Data) -> StageLineups:
        index = stream.read_short()
        total = stream.read_short()
        slots = [StageSlot.read(stream) for _ in range(total)]
        return StageLineups(index, slots)

    def write(self, stream: core.Data):
        stream.write_short(self.index)
        stream.write_short(len(self.slots))
        for slot in self.slots:
            slot.write(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "index": self.index,
            "slots": [slot.serialize() for slot in self.slots],
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> StageLineups:
        return StageLineups(
            data.get("index", 0),
            [StageSlot.deserialize(slot) for slot in data.get("slots", [])],
        )

    def __repr__(self):
        return f"StageLineups({self.index}, {self.slots})"

    def __str__(self):
        return self.__repr__()


class ClearedStageSlots:
    def __init__(self, lineups: list[StageLineups]):
        self.lineups = lineups

    @staticmethod
    def init() -> ClearedStageSlots:
        return ClearedStageSlots([])

    @staticmethod
    def read(stream: core.Data) -> ClearedStageSlots:
        total = stream.read_short()
        lineups = [StageLineups.read(stream) for _ in range(total)]
        return ClearedStageSlots(lineups)

    def write(self, stream: core.Data):
        stream.write_short(len(self.lineups))
        for lineup in self.lineups:
            lineup.write(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "lineups": [lineup.serialize() for lineup in self.lineups],
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> ClearedStageSlots:
        return ClearedStageSlots(
            [
                StageLineups.deserialize(lineup)
                for lineup in data.get("lineups", [])
            ],
        )

    def __repr__(self):
        return f"ClearedStageSlots({self.lineups})"

    def __str__(self):
        return self.__repr__()


class ClearedSlots:
    def __init__(
        self,
        cleared_slots: ClearedSlotsCat,
        cleared_stage_slots: ClearedStageSlots,
        unknown: dict[int, bool],
    ):
        self.cleared_slots = cleared_slots
        self.cleared_stage_slots = cleared_stage_slots
        self.unknown = unknown

    @staticmethod
    def init() -> ClearedSlots:
        return ClearedSlots(
            ClearedSlotsCat.init(),
            ClearedStageSlots.init(),
            {},
        )

    @staticmethod
    def read(stream: core.Data) -> ClearedSlots:
        cleared_slots = ClearedSlotsCat.read(stream)
        cleared_stage_slots = ClearedStageSlots.read(stream)
        length = stream.read_short()
        unknown = stream.read_short_bool_dict(length)
        return ClearedSlots(cleared_slots, cleared_stage_slots, unknown)

    def write(self, stream: core.Data):
        self.cleared_slots.write(stream)
        self.cleared_stage_slots.write(stream)
        stream.write_short(len(self.unknown))
        stream.write_short_bool_dict(self.unknown, write_length=False)

    def serialize(self) -> dict[str, Any]:
        return {
            "cleared_slots": self.cleared_slots.serialize(),
            "cleared_stage_slots": self.cleared_stage_slots.serialize(),
            "unknown": self.unknown,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> ClearedSlots:
        return ClearedSlots(
            ClearedSlotsCat.deserialize(data.get("cleared_slots", [])),
            ClearedStageSlots.deserialize(data.get("cleared_stage_slots", {})),
            data.get("unknown", {}),
        )

    def __repr__(self):
        return f"ClearedSlots({self.cleared_slots}, {self.cleared_stage_slots}, {self.unknown})"

    def __str__(self):
        return self.__repr__()



from __future__ import annotations
from bcsfe import core
from bcsfe.cli import color, dialog_creator
from typing import Union

ChaptersType = Union[
    "core.EventChapters",
    "core.GauntletChapters",
    "core.LegendQuestChapters",
    "core.ZeroLegendsChapters",
    "core.Chapters",
]


def get_total_maps(chapters: ChaptersType) -> int:
    if isinstance(chapters, core.EventChapters):
        return chapters.get_lengths()[1]
    return len(chapters.chapters)


def unclear_stage(
    chapters: ChaptersType,
    map: int,
    star: int,
    stage: int,
    type: int | None = None,
) -> bool:
    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")
        return chapters.unclear_stage(type, map, star, stage)
    else:
        return chapters.unclear_stage(map, star, stage)


def clear_stage(
    chapters: ChaptersType,
    map: int,
    star: int,
    stage: int,
    clear_amount: int = 1,
    overwrite_clear_progress: bool = False,
    type: int | None = None,
    ensure_cleared_only: bool = False,
) -> bool:
    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")

        return chapters.clear_stage(
            type, map, star, stage, clear_amount, overwrite_clear_progress
        )
    else:
        return chapters.clear_stage(
            map,
            star,
            stage,
            clear_amount,
            overwrite_clear_progress,
            ensure_cleared_only=ensure_cleared_only,
        )


def unclear_rest(
    chapters: ChaptersType,
    stages: list[int],
    stars: int,
    id: int,
    type: int | None = None,
):
    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")
        chapters.unclear_rest(stages, stars, id, type)
    else:
        chapters.unclear_rest(stages, stars, id)


def get_total_stars(
    map_option: core.MapOption,
    base_index: int,
    chapters: ChaptersType,
    id: int,
    type: int | None = None,
) -> int:

    max_stars = get_max_stars(chapters, id, type)

    map_option_stars = map_option.get_map(base_index + id)
    if map_option_stars is not None:
        return min(max_stars, map_option_stars.crown_count)
    return max_stars


def get_max_max_stars(
    map_option: core.MapOption,
    base_index: int,
    ids: list[int],
    chapters: ChaptersType,
    type: int | None = None,
) -> int:
    m = 0
    for id in ids:
        val = get_total_stars(map_option, base_index, chapters, id, type)
        if val > m:
            m = val

    return m


def get_max_stars(
    chapters: ChaptersType,
    id: int,
    type: int | None = None,
) -> int:

    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")
        max_stars = chapters.get_total_stars(type, id)
    else:
        max_stars = chapters.get_total_stars(id)

    return max_stars


def get_total_stages(
    chapters: ChaptersType, id: int, star: int, type: int | None = None
):
    if isinstance(chapters, core.EventChapters):
        if type is None:
            raise ValueError("Type must be specified for EventChapters!")
        total_stars = chapters.get_total_stages(type, id, star)
    else:
        total_stars = chapters.get_total_stages(id, star)

    return total_stars


def select_maps(
    save_file: core.SaveFile,
    chapters: ChaptersType,
    letter_code: str,
    base_index: int,
    no_r_prefix: bool = False,
) -> list[int] | None:
    map_names = core.MapNames(
        save_file, letter_code, no_r_prefix=no_r_prefix, base_index=base_index
    )
    names: dict[int, str | None] = {}
    for id, name in map_names.map_names.items():
        if id >= get_total_maps(chapters):
            continue
        names[id] = name

    return core.EventChapters.select_map_names(names)


def select_maps_stars(
    save_file: core.SaveFile,
    map_option: core.MapOption,
    chapters: ChaptersType,
    letter_code: str,
    base_index: int,
    type: int | None = None,
    no_r_prefix: bool = False,
) -> list[tuple[int, int]] | None:
    map_names = core.MapNames(
        save_file, letter_code, no_r_prefix=no_r_prefix, base_index=base_index
    )
    names: dict[int, str | None] = {}
    for id, name in map_names.map_names.items():
        if id >= get_total_maps(chapters):
            continue

        for star in range(get_total_stars(map_option, base_index, chapters, id, type)):
            names[id * 10 + star] = core.localize(
                "map_name_star", name=name, star=star + 1
            )

    ids = core.EventChapters.select_map_names(names)
    if ids is None:
        return None

    new_ids: list[tuple[int, int]] = []

    for id in ids:
        map_id = id // 10
        star_index = id % 10

        new_ids.append((map_id, star_index))

    return new_ids


def edit_chapters2_clear_count(
    save_file: core.SaveFile,
    chapters: ChaptersType,
    letter_code: str,
    base_index: int,
    type: int | None = None,
    no_r_prefix: bool = False,
):

    map_names = core.MapNames(
        save_file, letter_code, no_r_prefix=no_r_prefix, base_index=base_index
    )

    map_option = core.MapOption.from_save(save_file)
    if map_option is None:
        return None

    map_choices = select_maps_stars(
        save_file, map_option, chapters, letter_code, base_index, type, no_r_prefix
    )
    if map_choices is None:
        return None

    clear_all = edit_all_or_handle_ind(len(map_choices))
    if clear_all is None:
        return None

    if clear_all == 0:
        clear_count = core.EventChapters.ask_clear_amount()
        if clear_count is None:
            return None

        for local_map_id, star in map_choices:
            total_stages = get_total_stages(chapters, local_map_id, star, type)
            for stage in range(total_stages):
                clear_stage(chapters, local_map_id, star, stage, clear_count, type=type)
    else:
        for local_map_id, star in map_choices:
            print()
            core.EventChapters.print_current_chapter(
                core.localize(
                    "map_name_star",
                    star=star,
                    name=map_names.map_names.get(local_map_id),
                ),
                local_map_id,
            )
            clear_whole = dialog_creator.ChoiceInput.from_reduced(
                ["edit_whole_chapter", "edit_specific_stages"], dialog="edit_chapter_q"
            ).single_choice()
            if clear_whole is None:
                return None

            clear_whole -= 1

            if clear_whole == 0:
                clear_count = core.EventChapters.ask_clear_amount()
                if clear_count is None:
                    return None

                for stage in range(
                    get_total_stages(chapters, local_map_id, star, type)
                ):
                    clear_stage(
                        chapters, local_map_id, star, stage, clear_count, type=type
                    )
            else:
                stage_ids = core.EventChapters.ask_stages(map_names, local_map_id)

                if stage_ids is None:
                    return None

                all_selected_stages = dialog_creator.ChoiceInput.from_reduced(
                    ["each_stage_individually", "stage_all_at_once"],
                    dialog="set_clear_count_stage_q",
                ).single_choice()
                if all_selected_stages is None:
                    return None

                all_selected_stages -= 1

                stage_names = core.EventChapters.get_stage_names(
                    map_names, local_map_id
                )
                if stage_names is None:
                    stage_names = []
                if all_selected_stages == 0:
                    for stage in stage_ids:
                        print()
                        if stage < len(stage_names):
                            stage_name = stage_names[stage]
                        else:
                            stage_name = None
                        core.EventChapters.print_current_stage(stage_name, stage)
                        clear_count = core.EventChapters.ask_clear_amount()
                        if clear_count is None:
                            return None
                        clear_stage(
                            chapters, local_map_id, star, stage, clear_count, type=type
                        )
                else:
                    clear_count = core.EventChapters.ask_clear_amount()
                    if clear_count is None:
                        return None
                    for stage in stage_ids:
                        clear_stage(
                            chapters, local_map_id, star, stage, clear_count, type=type
                        )


def clear_all_or_handle_ind(map_choices_len: int) -> int | None:
    if map_choices_len <= 1:
        clear_all = 1
    else:
        clear_all = dialog_creator.ChoiceInput.from_reduced(
            ["clear_all", "handle_individually"], dialog="clear_chapters_q"
        ).single_choice()
        if clear_all is None:
            return None

        clear_all -= 1

    return clear_all


def unclear_all_or_handle_ind(map_choices_len: int) -> int | None:
    if map_choices_len <= 1:
        clear_all = 1
    else:
        clear_all = dialog_creator.ChoiceInput.from_reduced(
            ["unclear_all", "handle_individually"], dialog="unclear_chapters_q"
        ).single_choice()
        if clear_all is None:
            return None

        clear_all -= 1

    return clear_all


def edit_all_or_handle_ind(map_choices_len: int) -> int | None:
    if map_choices_len <= 1:
        clear_all = 1
    else:
        clear_all = dialog_creator.ChoiceInput.from_reduced(
            ["edit_map_all", "handle_individually"], dialog="edit_chapters_q_all"
        ).single_choice()
        if clear_all is None:
            return None

        clear_all -= 1

    return clear_all


def edit_chapters2_progress(
    save_file: core.SaveFile,
    chapters: ChaptersType,
    letter_code: str,
    base_index: int,
    type: int | None = None,
    no_r_prefix: bool = False,
    allow_unclear: bool = False,
):
    map_names = core.MapNames(
        save_file, letter_code, no_r_prefix=no_r_prefix, base_index=base_index
    )

    map_choices = select_maps(save_file, chapters, letter_code, base_index, no_r_prefix)
    if map_choices is None:
        return None

    clear_all = clear_all_or_handle_ind(len(map_choices))
    if clear_all is None:
        return None

    map_option = core.MapOption.from_save(save_file)
    if map_option is None:
        return None

    if clear_all == 0:
        max_stars = get_max_max_stars(
            map_option, base_index, map_choices, chapters, type
        )
        if allow_unclear:
            stars = core.EventChapters.ask_stars_unclear(max_stars, "max_stars")
        else:
            stars = core.EventChapters.ask_stars(max_stars, "max_stars")
        if stars is None:
            return None
        for local_map_id in map_choices:
            unclear_rest(
                chapters,
                [0],
                max(0, stars - 1),
                local_map_id,
                type,
            )
            for star in range(stars):
                total_stages = get_total_stages(chapters, local_map_id, star, type)
                for stage in range(total_stages):
                    clear_stage(
                        chapters,
                        local_map_id,
                        star,
                        stage,
                        type=type,
                        ensure_cleared_only=True,
                    )

        return map_choices

    for local_map_id in map_choices:
        name = map_names.map_names.get(local_map_id)
        core.EventChapters.print_current_chapter(name, local_map_id)
        clear_whole = dialog_creator.ChoiceInput.from_reduced(
            ["clear_whole_chapter", "clear_to_specific_stage"], dialog="clear_whole_q"
        ).single_choice()
        if clear_whole is None:
            return None

        clear_whole -= 1

        if clear_whole == 0:
            max_stars = get_total_stars(
                map_option, base_index, chapters, local_map_id, type
            )
            if allow_unclear:
                stars = core.EventChapters.ask_stars_unclear(max_stars)
            else:
                stars = core.EventChapters.ask_stars(max_stars)
            if stars is None:
                return None

            unclear_rest(
                chapters,
                [0],
                max(stars - 1, 0),
                local_map_id,
                type,
            )

            for star in range(stars):
                total_stages = get_total_stages(chapters, local_map_id, star, type)
                for stage in range(total_stages):
                    clear_stage(
                        chapters,
                        local_map_id,
                        star,
                        stage,
                        type=type,
                        ensure_cleared_only=True,
                    )

        else:
            stage_names = map_names.stage_names.get(local_map_id)
            stage_names = [
                stage_name
                for stage_name in stage_names or []
                if stage_name and stage_name != "＠"
            ]
            stage_id = core.EventChapters.ask_stages_stage_names_one(stage_names)
            if stage_id is None:
                return None

            max_stars = get_total_stars(
                map_option, base_index, chapters, local_map_id, type
            )

            if allow_unclear:
                stars = core.EventChapters.ask_stars_unclear(max_stars)
            else:
                stars = core.EventChapters.ask_stars(max_stars)
            if stars is None:
                return None

            unclear_rest(
                chapters, list(range(stage_id)), max(stars - 1, 0), local_map_id, type
            )

            for star in range(stars - 1):
                total_stages = get_total_stages(chapters, local_map_id, star, type)
                for stage in range(total_stages):
                    clear_stage(
                        chapters,
                        local_map_id,
                        star,
                        stage,
                        type=type,
                        ensure_cleared_only=True,
                    )

            for stage in range(stage_id + 1):
                clear_stage(
                    chapters,
                    local_map_id,
                    stars - 1,
                    stage,
                    type=type,
                    ensure_cleared_only=True,
                )


def edit_chapters(
    save_file: core.SaveFile,
    chapters: ChaptersType,
    letter_code: str,
    base_index: int,
    type: int | None = None,
    no_r_prefix: bool = False,
) -> dict[int, bool] | None:
    while True:
        choice = dialog_creator.ChoiceInput.from_reduced(
            [
                "edit_progress_clear",
                "edit_progress_unclear",
                "edit_clear_counts",
                "finish",
            ],
            dialog="edit_chapters_q",
        ).single_choice()
        if choice is None:
            return None
        choice -= 1

        if choice == 0:
            edit_chapters2_progress(
                save_file, chapters, letter_code, base_index, type, no_r_prefix
            )
        elif choice == 1:
            edit_chapters2_progress(
                save_file,
                chapters,
                letter_code,
                base_index,
                type,
                no_r_prefix,
                allow_unclear=True,
            )
        elif choice == 2:
            edit_chapters2_clear_count(
                save_file, chapters, letter_code, base_index, type, no_r_prefix
            )
        else:
            break
        color.ColoredText.localize("map_chapters_edited")
    color.ColoredText.localize("map_chapters_edited")

    return None



from __future__ import annotations
import enum
from bcsfe import core


class GatyaItemNames:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.names = self.__get_names()

    def __get_names(self) -> list[str] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("resLocal", "GatyaitemName.csv")
        if data is None:
            return None
        csv = core.CSV(
            data, core.Delimeter.from_country_code_res(self.save_file.cc)
        )
        names: list[str] = []
        for line in csv:
            names.append(line[0].to_str())

        return names

    def get_name(self, index: int) -> str | None:
        if self.names is None:
            return None
        try:
            return self.names[index]
        except IndexError:
            return core.core_data.local_manager.get_key(
                "gatya_item_unknown_name", index=index
            )


class GatyaItemBuyItem:
    def __init__(
        self,
        id: int,
        rarity: int,
        reflect_or_storage: bool,
        price: int,
        stage_drop_id: int,
        quantity: int,
        server_id: int,
        category: int,
        index: int,
        src_item_id: int,
        main_menu_type: int,
        gatya_ticket_id: int,
        comment: str,
    ):
        self.id = id
        self.rarity = rarity
        self.reflect_or_storage = reflect_or_storage
        self.price = price
        self.stage_drop_id = stage_drop_id
        self.quantity = quantity
        self.server_id = server_id
        self.category = category
        self.index = index
        self.src_item_id = src_item_id
        self.main_menu_type = main_menu_type
        self.gatya_ticket_id = gatya_ticket_id
        self.comment = comment

class GatyaItemCategory(enum.Enum):
    MISC = 0
    EVENT_TICKETS = 1
    SPECIAL_SKILLS = 2
    BATTLE_ITEMS = 3
    EVOLVE_ITEMS = 4
    CATSEYES = 5
    CATAMINS = 6
    BASE_MATERIALS = 7
    LUCKY_TICKETS_1 = 8
    ENDLESS_ITEMS = 9
    LUCKY_TICKETS_2 = 10
    LABYRINTH_MEDALS = 11
    TREASURE_CHESTS = 12

class GatyaItemBuy:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.buy = self.get_buy()

    def get_buy(self) -> list[GatyaItemBuyItem] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "Gatyaitembuy.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        buy: list[GatyaItemBuyItem] = []
        for i, line in enumerate(csv.lines[1:]):
            try:
                buy.append(
                    GatyaItemBuyItem(
                        i,
                        line[0].to_int(),
                        line[1].to_bool(),
                        line[2].to_int(),
                        line[3].to_int(),
                        line[4].to_int(),
                        line[5].to_int(),
                        line[6].to_int(),
                        line[7].to_int(),
                        line[8].to_int(),
                        line[9].to_int(),
                        line[10].to_int(),
                        line[11].to_str(),
                    )
                )
            except IndexError:
                pass

        return buy

    def sort_by_index(self, items: list[GatyaItemBuyItem]):
        items.sort(key=lambda x: x.index)
        return items

    def get_by_category(self, category: int | GatyaItemCategory) -> list[GatyaItemBuyItem] | None:
        if self.buy is None:
            return None
        if isinstance(category, GatyaItemCategory):
            category = category.value
        return self.sort_by_index(
            [item for item in self.buy if item.category == category]
        )

    def get_names_by_category(self, category: int | GatyaItemCategory) -> list[tuple[GatyaItemBuyItem, str | None]] | None:
        items = self.get_by_category(category)
        if items is None:
            return None

        names = GatyaItemNames(self.save_file)

        return [(item, names.get_name(item.id)) for item in items]

    def get(self, item_id: int) -> GatyaItemBuyItem | None:
        if self.buy is None:
            return None
        if item_id < 0 or item_id >= len(self.buy):
            return None

        return self.buy[item_id]

    def get_by_server_id(self, server_id: int) -> GatyaItemBuyItem | None:
        if self.buy is None:
            return None
        for item in self.buy:
            if item.server_id == server_id:
                return item

        return None



from __future__ import annotations
from bcsfe import cli, core
from bcsfe.core.game.catbase.gatya import GatyaEventType
from bcsfe.core.server.event_data import split_hhmm, split_yyyymmdd


class EventTickets:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.gatya_item_buy = core.core_data.get_gatya_item_buy(self.save_file)
        self.gatya_item_names = core.core_data.get_gatya_item_names(self.save_file)
        self.gatya_option_n = core.GatyaDataOption.read(
            self.save_file, GatyaEventType.NORMAL
        )
        self.gatya_option_r = core.GatyaDataOption.read(
            self.save_file, GatyaEventType.RARE
        )
        self.gatya_option_e = core.GatyaDataOption.read(
            self.save_file, GatyaEventType.EVENT
        )

        cli.color.ColoredText.localize("downloading_gatya_data")
        temp_save_file = core.SaveFile(cc=save_file.cc, gv=save_file.game_version)
        gatya_event_data = core.ServerHandler(temp_save_file).download_gatya_data()

        if gatya_event_data is None:
            cli.color.ColoredText.localize("download_gatya_data_fail")
            self.gatya_event_data = None
        else:
            cli.color.ColoredText.localize("download_gatya_data_success")
            self.gatya_event_data = core.ServerGatyaData.from_data(gatya_event_data)

    @staticmethod
    def edit(save_file: core.SaveFile):
        event_tickets = EventTickets(save_file)

        if event_tickets.gatya_event_data is None:
            return

        event_ticket_items: list[
            tuple[
                core.ServerGatyaDataItem, core.ServerGatyaDataSet, core.GatyaItemBuyItem
            ]
        ] = []

        if (
            event_tickets.gatya_option_n is None
            or event_tickets.gatya_option_r is None
            or event_tickets.gatya_option_e is None
        ):
            return

        for item in event_tickets.gatya_event_data.items:
            for gset in item.sets:
                if gset.number == -1:
                    continue

                gset_opt = None

                if item.get_normal_flag():
                    gset_opt = event_tickets.gatya_option_n.get(gset.number)
                elif item.get_rare_flag():
                    gset_opt = event_tickets.gatya_option_r.get(gset.number)
                elif item.get_collab_flag():
                    gset_opt = event_tickets.gatya_option_e.get(gset.number)

                if gset_opt is None:
                    continue

                gatya_item = event_tickets.gatya_item_buy.get(gset_opt.ticket_item_id)
                if gatya_item is None:
                    continue

                category = gatya_item.category
                if category in [
                    core.GatyaItemCategory.EVENT_TICKETS.value,
                    core.GatyaItemCategory.LUCKY_TICKETS_1.value,
                    core.GatyaItemCategory.LUCKY_TICKETS_2.value,
                ]:
                    event_ticket_items.append((item, gset, gatya_item))

        event_names: list[str] = []
        values: list[int] = []

        for event_item, gset, gatya_item in event_ticket_items:
            start_y, start_m, start_d = split_yyyymmdd(event_item.filter.start_yyyymmdd)
            start_h, start_min = split_hhmm(event_item.filter.start_hhmm)
            end_y, end_m, end_d = split_yyyymmdd(event_item.filter.end_yyyymmdd)
            end_h, end_min = split_hhmm(event_item.filter.end_hhmm)
            time_str = f"{start_y}-{start_m:02}-{start_d:02} {start_h:02}:{start_min:02} -> {end_y}-{end_m:02}-{end_d:02} {end_h:02}:{end_min:02}"
            event_message = gset.message.replace("<br>", "\n")

            base_msg = f"{time_str}"
            item_name = event_tickets.gatya_item_names.get_name(gatya_item.id)
            if item_name is not None:
                base_msg += f" - {item_name}"

            if event_message:
                base_msg += f" - {event_message}"

            current_amount = event_tickets.get_ticket(gatya_item.id)

            if current_amount is not None:
                event_names.append(base_msg)
                values.append(current_amount)

        values = cli.dialog_creator.MultiEditor.from_reduced(
            "event_tickets",
            event_names,
            ints=values,
            max_values=core.core_data.max_value_manager.get("event_tickets"),
            group_name_localized=True,
        ).edit()

        for (event_item, gset, gatya_item), value in zip(event_ticket_items, values):
            event_tickets.edit_ticket(gatya_item.id, value)

    def get_ticket(self, item_id: int) -> int | None:
        item = self.gatya_item_buy.get(item_id)
        if item is None:
            return

        if item.category == core.GatyaItemCategory.EVENT_TICKETS.value:
            if item.index < len(self.save_file.event_capsules):
                return self.save_file.event_capsules[item.index]
        if item.category == core.GatyaItemCategory.LUCKY_TICKETS_1.value:
            if item.index < len(self.save_file.lucky_tickets):
                return self.save_file.lucky_tickets[item.index]
        if item.category == core.GatyaItemCategory.LUCKY_TICKETS_2.value:
            if item.index < len(self.save_file.event_capsules_2):
                return self.save_file.event_capsules_2[item.index]

        return None

    def edit_ticket(self, item_id: int, amount: int):
        item = self.gatya_item_buy.get(item_id)
        if item is None:
            return

        if item.category == core.GatyaItemCategory.EVENT_TICKETS.value:
            if item.index < len(self.save_file.event_capsules):
                self.save_file.event_capsules[item.index] = amount
        if item.category == core.GatyaItemCategory.LUCKY_TICKETS_1.value:
            if item.index < len(self.save_file.lucky_tickets):
                self.save_file.lucky_tickets[item.index] = amount
        if item.category == core.GatyaItemCategory.LUCKY_TICKETS_2.value:
            if item.index < len(self.save_file.event_capsules_2):
                self.save_file.event_capsules_2[item.index] = amount



from __future__ import annotations
from bcsfe import core


class Popup:
    def __init__(self, seen: bool):
        self.seen = seen

    @staticmethod
    def init() -> Popup:
        return Popup(False)

    @staticmethod
    def read(stream: core.Data) -> Popup:
        seen = stream.read_bool()
        return Popup(seen)

    def write(self, stream: core.Data):
        stream.write_bool(self.seen)

    def serialize(self) -> bool:
        return self.seen

    @staticmethod
    def deserialize(data: bool) -> Popup:
        return Popup(data)

    def __repr__(self) -> str:
        return f"Popup(seen={self.seen!r})"

    def __str__(self) -> str:
        return self.__repr__()


class UnlockPopups:
    def __init__(self, popups: dict[int, Popup]):
        self.popups = popups

    @staticmethod
    def init() -> UnlockPopups:
        return UnlockPopups({})

    @staticmethod
    def read(stream: core.Data) -> UnlockPopups:
        total = stream.read_int()
        popups: dict[int, Popup] = {}
        for _ in range(total):
            key = stream.read_int()
            popups[key] = Popup.read(stream)
        return UnlockPopups(popups)

    def write(self, stream: core.Data):
        stream.write_int(len(self.popups))
        for key, popup in self.popups.items():
            stream.write_int(key)
            popup.write(stream)

    def serialize(self) -> dict[int, bool]:
        return {key: popup.serialize() for key, popup in self.popups.items()}

    @staticmethod
    def deserialize(data: dict[int, bool]) -> UnlockPopups:
        return UnlockPopups(
            {int(key): Popup.deserialize(popup) for key, popup in data.items()}
        )

    def __repr__(self) -> str:
        return f"Popups(popups={self.popups!r})"

    def __str__(self) -> str:
        return self.__repr__()


class UnlockPopupLine:
    def __init__(
        self,
        popup_id: int,
        enabled: bool,
        conditions: int,
        stage: int,
        map_conditions: int,
        user_rank: int,
        get_char_id1: int,
        get_char_id2: int,
        os_id: int,
        unlock_eye_1_id: int,
        add_level1: int,
        unlock_eye_2_id: int,
        add_level2: int,
        unlock_plus_id: int,
        add_level: int,
        skill_id: int,
        item_id: int,
        num: int,
        help_enabled: bool,
    ):
        self.popup_id = popup_id
        self.enabled = enabled
        self.conditions = conditions
        self.stage = stage
        self.map_conditions = map_conditions
        self.user_rank = user_rank
        self.get_char_id1 = get_char_id1
        self.get_char_id2 = get_char_id2
        self.os_id = os_id
        self.unlock_eye_1_id = unlock_eye_1_id
        self.add_level1 = add_level1
        self.unlock_eye_2_id = unlock_eye_2_id
        self.add_level2 = add_level2
        self.unlock_plus_id = unlock_plus_id
        self.add_level = add_level
        self.skill_id = skill_id
        self.item_id = item_id
        self.num = num
        self.help_enabled = help_enabled

    @staticmethod
    def from_csv_row(row: core.Row) -> UnlockPopupLine:
        return UnlockPopupLine(
            row.next_int(),
            row.next_bool(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_bool(),
        )


class UnlockPopupData:
    def __init__(self, popups: list[UnlockPopupLine]):
        self.popups = popups

    @staticmethod
    def from_csv(csv: core.CSV) -> UnlockPopupData:
        popups: list[UnlockPopupLine] = []
        for line in csv.lines[1:]:
            popups.append(UnlockPopupLine.from_csv_row(line))

        return UnlockPopupData(popups)

    @staticmethod
    def from_save(save_file: core.SaveFile) -> UnlockPopupData | None:
        gdg = core.core_data.get_game_data_getter(save_file)
        data = gdg.download("DataLocal", "unlockPopup.tsv")
        if data is None:
            return None

        csv = core.CSV(data, "\t")

        return UnlockPopupData.from_csv(csv)



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import dialog_creator


class EquipSlot:
    def __init__(self, cat_id: int):
        self.cat_id = cat_id

    @staticmethod
    def read(stream: core.Data) -> EquipSlot:
        return EquipSlot(stream.read_int())

    def write(self, stream: core.Data):
        stream.write_int(self.cat_id)

    def serialize(self) -> int:
        return self.cat_id

    @staticmethod
    def deserialize(data: int) -> EquipSlot:
        return EquipSlot(data)

    def __repr__(self):
        return f"EquipSlot({self.cat_id})"

    def __str__(self):
        return f"EquipSlot({self.cat_id})"


class EquipSlots:
    def __init__(self, slots: list[EquipSlot]):
        self.slots = slots
        self.name = ""

    @staticmethod
    def read(stream: core.Data) -> EquipSlots:
        length = 10
        slots = [EquipSlot.read(stream) for _ in range(length)]
        return EquipSlots(slots)

    @staticmethod
    def init() -> EquipSlots:
        length = 10
        slots = [EquipSlot(-1) for _ in range(length)]
        return EquipSlots(slots)

    def write(self, stream: core.Data):
        for slot in self.slots:
            slot.write(stream)

    def read_name(self, stream: core.Data):
        length = stream.read_int()
        try:
            self.name = stream.read_string(length)
        except UnicodeDecodeError:
            stream.pos -= length
            self.name = stream.read_utf8_string_by_char_length(length)

    def write_name(self, stream: core.Data):
        stream.write_string(self.name)

    def serialize(self) -> dict[str, Any]:
        return {
            "slots": [slot.serialize() for slot in self.slots],
            "name": self.name,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> EquipSlots:
        slots = EquipSlots(
            [EquipSlot.deserialize(slot) for slot in data.get("slots", [])]
        )
        slots.name = data.get("name")
        return slots

    def __repr__(self):
        return f"EquipSlots({self.slots}, {self.name})"

    def __str__(self):
        return self.__repr__()


class LineUps:
    def __init__(self, slots: list[EquipSlots], total_slots: int = 15):
        self.slots = slots
        self.selected_slot = 0
        self.unlocked_slots = 0
        self.slot_names_length = total_slots

    @staticmethod
    def init(gv: core.GameVersion) -> LineUps:
        if gv < 90700:
            length = 10
        else:
            length = 15
        slots = [EquipSlots.init() for _ in range(length)]
        return LineUps(slots, length)

    @staticmethod
    def read(stream: core.Data, gv: core.GameVersion) -> LineUps:
        if gv < 90700:
            length = 10
        else:
            length = stream.read_byte()
        slots = [EquipSlots.read(stream) for _ in range(length)]
        return LineUps(slots)

    def write(self, stream: core.Data, gv: core.GameVersion):
        if gv >= 90700:
            stream.write_byte(len(self.slots))
            length = len(self.slots)
        else:
            length = 10
        if length > len(self.slots):
            self.slots += [EquipSlots.init() for _ in range(length)]
        else:
            self.slots = self.slots[:length]
        for slot in self.slots:
            slot.write(stream)

    def read_2(self, stream: core.Data, gv: core.GameVersion):
        self.selected_slot = stream.read_int()
        if gv < 90700:
            unlocked_slots_l = stream.read_bool_list(10)
            unlocked_slots = sum(unlocked_slots_l)
        else:
            unlocked_slots = stream.read_byte()
        self.unlocked_slots = unlocked_slots

    def write_2(self, stream: core.Data, gv: core.GameVersion):
        stream.write_int(self.selected_slot)
        if gv < 90700:
            unlocked_slots_l = [False] * 10
            unlocked_slots = min(self.unlocked_slots, 10)
            for i in range(unlocked_slots):
                unlocked_slots_l[i] = True
            stream.write_bool_list(unlocked_slots_l, write_length=False)
        else:
            stream.write_byte(self.unlocked_slots)

    def read_slot_names(self, stream: core.Data, gv: core.GameVersion):
        if gv >= 110600:
            total_slots = stream.read_byte()
        else:
            total_slots = 15
        for i in range(total_slots):
            try:
                self.slots[i].read_name(stream)
            except IndexError:
                slot = EquipSlots.init()
                slot.read_name(stream)
                self.slots.append(slot)

        self.slot_names_length = total_slots

    def write_slot_names(self, stream: core.Data, gv: core.GameVersion):
        if gv >= 110600:
            stream.write_byte(self.slot_names_length)
        for i in range(self.slot_names_length):
            try:
                self.slots[i].write_name(stream)
            except IndexError:
                slot = EquipSlots.init()
                slot.write_name(stream)
                self.slots.append(slot)

    def serialize(self) -> dict[str, Any]:
        return {
            "slots": [slot.serialize() for slot in self.slots],
            "selected_slot": self.selected_slot,
            "unlocked_slots": self.unlocked_slots,
            "slot_names_length": self.slot_names_length,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> LineUps:
        line_ups = LineUps(
            [EquipSlots.deserialize(slot) for slot in data.get("slots", [])]
        )
        line_ups.selected_slot = data.get("selected_slot", 0)
        line_ups.unlocked_slots = data.get("unlocked_slots", 0)
        line_ups.slot_names_length = data.get("slot_names_length", 0)
        return line_ups

    def __repr__(self):
        return f"LineUps({self.slots}, {self.selected_slot}, {self.unlocked_slots})"

    def __str__(self):
        return self.__repr__()

    def edit_unlocked_slots(self):
        self.unlocked_slots = dialog_creator.SingleEditor(
            "unlocked_slots",
            self.unlocked_slots,
            self.slot_names_length,
            localized_item=True,
            remove_alias=True,
        ).edit()



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe import cli
import yaml


class YamlFile:
    def __init__(self, path: core.Path, print_err: bool = True):
        self.path = path
        self.yaml: dict[str, Any] = {}
        if self.path.exists():
            self.data = path.read()
            try:
                yml = yaml.safe_load(self.data.data)
                if not isinstance(yml, dict):
                    self.yaml = {}
                    self.save(print_err)
                else:
                    self.yaml = yml
            except yaml.YAMLError:
                self.yaml = {}
                self.save(print_err)
        else:
            self.yaml = {}
            self.save(print_err)

    def save(self, print_err: bool = True) -> None:
        self.path.parent().generate_dirs()

        try:
            with open(self.path.path, "w", encoding="utf-8") as f:
                yaml.dump(self.yaml, f)
        except FileNotFoundError:
            if print_err:
                cli.color.ColoredText.localize("yaml_create_error", path=self.path.path)

    def __getitem__(self, key: str) -> Any:
        return self.yaml[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.yaml[key] = value

    def __delitem__(self, key: str) -> None:
        del self.yaml[key]

    def __contains__(self, key: str) -> bool:
        return key in self.yaml

    def __iter__(self):
        return iter(self.yaml)

    def __len__(self) -> int:
        return len(self.yaml)

    def __repr__(self) -> str:
        return self.yaml.__repr__()

    def __str__(self) -> str:
        return self.yaml.__str__()

    def get(self, key: str) -> Any:
        return self.yaml.get(key)

    def remove(self) -> None:
        self.path.remove()
        self.yaml = {}



from __future__ import annotations

import requests

from bcsfe import core


class MultiPartFile:
    def __init__(self, content: bytes, content_type: str, filename: str | None = None):
        self.content = content
        self.content_type = content_type
        self.filename = filename


class MultipartForm:
    def __init__(self):
        self.data: dict[str, MultiPartFile] = {}

    def into_files(
        self,
    ) -> dict[str, tuple[str | None, bytes, str]]:
        out = {}
        for name, data in self.data.items():
            out[name] = (data.filename, data.content, data.content_type)

        return out

    def add_key(
        self, key: str, content: bytes, content_type: str, filename: str | None = None
    ):
        self.data[key] = MultiPartFile(content, content_type, filename)

    def get_all_type(self, content_type: str) -> str:
        data = ""
        for key, file in self.data.items():
            if file.content_type == content_type:
                content = file.content.decode("utf-8", errors="ignore")
                data += f"key: {key}, data: {content}\n"

        return data


class RequestHandler:
    """Handles HTTP requests."""

    def __init__(
        self,
        url: str,
        headers: dict[str, str] | None = None,
        data: core.Data | None = None,
        form: MultipartForm | None = None,
    ):
        """Initializes a new instance of the RequestHandler class.

        Args:
            url (str): URL to request.
            headers (dict[str, str] | None, optional): Headers to send with the request. Defaults to None.
            data (core.Data | None, optional): Data to send with the request. Defaults to None.
        """
        if data is None:
            data = core.Data()
        self.url = url
        self.headers = headers
        self.data = data
        self.form = form

    def get(
        self,
        stream: bool = False,
        no_timeout: bool = False,
    ) -> requests.Response | None:
        """Sends a GET request.

        Returns:
            requests.Response: Response from the server.
        """
        try:
            return requests.get(
                self.url,
                headers=self.headers,
                timeout=(
                    None
                    if no_timeout
                    else core.core_data.config.get_int(
                        core.ConfigKey.MAX_REQUEST_TIMEOUT
                    )
                ),
                stream=stream,
                files=None if self.form is None else self.form.into_files(),
            )
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
            return None

    def post(self, no_timeout: bool = False) -> requests.Response | None:
        """Sends a POST request.

        Returns:
            requests.Response: Response from the server.
        """
        try:
            return requests.post(
                self.url,
                headers=self.headers,
                data=self.data.data,
                timeout=(
                    None
                    if no_timeout
                    else core.core_data.config.get_int(
                        core.ConfigKey.MAX_REQUEST_TIMEOUT
                    )
                ),
                files=None if self.form is None else self.form.into_files(),
            )
        except requests.exceptions.ConnectionError:
            return None



from __future__ import annotations

"""Main class for the CLI."""

import sys
import traceback
from typing import Any, NoReturn
from bcsfe.cli import (
    file_dialog,
    color,
    feature_handler,
    save_management,
    dialog_creator,
)
from bcsfe import core


class Main:
    """Main class for the CLI."""

    def __init__(self):
        self.save_file = None
        self.exit = False
        self.save_path = None
        self.fh = None

    def wipe_temp_save(self):
        """Wipe the temp save."""
        core.SaveFile.get_temp_path().remove()

    def main(self, input_path: str | None = None):
        """Main function for the CLI."""
        self.wipe_temp_save()
        core.GameDataGetter.delete_old_versions(5)
        self.check_update()
        print()
        self.print_start_text()
        while not self.exit:
            stop = self.load_save_options(input_path)
            if stop:
                break

    def check_update(self):
        """Check for updates."""

        updater = core.Updater()
        has_pre_release = updater.has_enabled_pre_release()
        local_version = updater.get_local_version()
        latest_version = updater.get_latest_version(has_pre_release)

        if latest_version is None:
            color.ColoredText.localize("update_check_fail")
            return

        color.ColoredText.localize(
            "version_line",
            local_version=local_version,
            latest_version=latest_version,
        )

        is_local_beta = "b" in local_version
        is_latest_beta = "b" in latest_version

        local_no_beta = local_version.split("b")[0]
        latest_no_beta = latest_version.split("b")[0]

        if latest_no_beta > local_no_beta:
            update_needed = True
        elif latest_no_beta < local_no_beta:
            update_needed = False
        else:
            if latest_version == local_version:
                update_needed = False
            else:
                if is_local_beta and is_latest_beta:
                    update_needed = latest_version > local_version
                elif is_local_beta:
                    update_needed = True
                else:
                    update_needed = False

        show_message = core.core_data.config.get(core.ConfigKey.SHOW_UPDATE_MESSAGE)
        if not show_message:
            update_needed = False

        if update_needed:
            update = dialog_creator.YesNoInput(True).get_input_once(
                "update_available", {"latest_version": latest_version}
            )
            if update is None:
                return

            if update:
                if updater.update(latest_version):
                    color.ColoredText.localize("update_success")
                else:
                    color.ColoredText.localize("update_fail")
                sys.exit()
            else:
                disable_message = dialog_creator.YesNoInput(False).get_input_once(
                    "disable_update_message"
                )
                if disable_message is None:
                    return

                core.core_data.config.set(
                    core.ConfigKey.SHOW_UPDATE_MESSAGE, not disable_message
                )

    def print_start_text(self):
        external_theme = core.ExternalThemeManager.get_external_theme_config()
        external_locale = core.ExternalLocaleManager.get_external_locale_config()
        if external_theme is None:
            theme_text = core.core_data.local_manager.get_key(
                "theme_text",
                theme_path=core.ThemeHandler.get_theme_path(
                    core.core_data.theme_manager.theme_code
                ),
                theme_version=core.core_data.theme_manager.get_version(),
                theme_author=core.core_data.theme_manager.get_author(),
                theme_name=core.core_data.theme_manager.get_name(),
                escape=False,
            )
        else:
            theme_text = core.core_data.local_manager.get_key(
                "theme_text",
                theme_name=external_theme.name,
                theme_version=external_theme.version,
                theme_author=external_theme.author,
                theme_path=core.ThemeHandler.get_theme_path(
                    external_theme.get_full_name()
                ),
                escape=False,
            )
        if external_locale is None:
            authors = core.core_data.local_manager.authors
            locale_text = core.core_data.local_manager.get_key(
                "default_locale_text_authors",
                path=core.core_data.local_manager.path,
                authors=", ".join(authors),
                name=core.core_data.local_manager.name,
                escape=False,
            )
        else:
            locale_text = core.core_data.local_manager.get_key(
                "locale_text",
                locale_name=external_locale.name,
                locale_version=external_locale.version,
                locale_author=external_locale.author,
                locale_path=core.LocalManager.get_locale_folder(
                    external_locale.get_full_name()
                ),
                escape=False,
            )
        color.ColoredText.localize(
            "welcome",
            config_path=core.core_data.config.get_config_path(),
            locale_text=locale_text,
            theme_text=theme_text,
            escape=False,
        )
        print()

    def load_save_options(self, input_path: str | None = None):
        """Load save options."""
        save_file, stop = save_management.SaveManagement.select_save(True, input_path)
        if save_file is None:
            return stop
        self.save_file = save_file

        color.ColoredText.localize(
            "current_save",
            inquiry_code=save_file.inquiry_code[:4]
            + "***"
            + save_file.inquiry_code[-2:],
            gv=save_file.game_version,
            cc=save_file.cc,
        )

        self.feature_handler()
        return False

    def feature_handler(self):
        """Run the feature handler."""
        if self.save_file is None:
            return
        self.fh = feature_handler.FeatureHandler(self.save_file)
        self.fh.select_features_run()

    @staticmethod
    def save_save_dialog(save_file: core.SaveFile) -> core.Path | None:
        """Save save file dialog.

        Args:
            save_file (core.SaveFile): Save file to save.

        Returns:
            core.Path: Path to save file.
        """
        path = file_dialog.FileDialog().save_file(
            "save_save_dialog",
            initialdir=core.SaveFile.get_saves_path().to_str(),
            initialfile="SAVE_DATA",
        )
        if path is None:
            return None
        path = core.Path(path)
        path.parent().generate_dirs()
        save_file.save_path = path
        return path

    @staticmethod
    def save_json_dialog(json_data: dict[str, Any]) -> core.Path | None:
        """Save json file dialog.

        Args:
            json_data (dict): Json data to save.

        Returns:
            core.Path: Path to save file.
        """
        path = file_dialog.FileDialog().save_file(
            "save_json_dialog",
            initialfile="SAVE_DATA.json",
            initialdir=core.SaveFile.get_saves_path().to_str(),
        )
        if path is None:
            return None
        path = core.Path(path)
        path.parent().generate_dirs()
        core.JsonFile.from_object(json_data).to_data().to_file(path)
        return path

    @staticmethod
    def load_save_file() -> core.Path | None:
        """Load save file from file dialog.

        Returns:
            core.Path: Path to save file.
        """
        path = file_dialog.FileDialog().get_file(
            "select_save_file",
            initialdir=core.SaveFile.get_saves_path().to_str(),
            initialfile="SAVE_DATA",
            ignore_json=True,
        )
        if path is None:
            return None
        path = core.Path(path)
        return path

    @staticmethod
    def load_save_data_json() -> tuple[core.Path, core.CountryCode] | None:
        """Load save data from json file.

        Returns:
            core.Path: Path to save file.
        """
        path = file_dialog.FileDialog().get_file(
            "load_save_data_json",
            initialfile="SAVE_DATA.json",
            initialdir=core.SaveFile.get_saves_path().to_str(),
        )
        if path is None:
            return None
        path = core.Path(path)
        if not path.exists():
            return None
        try:
            json_data = core.JsonFile.from_data(path.read()).to_object()
        except core.JSONDecodeError:
            color.ColoredText.localize(
                "load_json_fail", error=core.core_data.logger.get_traceback()
            )
            return None
        try:
            save_file = core.SaveFile.from_dict(json_data)
        except core.SaveError:
            color.ColoredText.localize(
                "load_json_fail", error=core.core_data.logger.get_traceback()
            )
            return None
        path = Main.save_save_dialog(save_file)
        if path is None:
            return None
        save_file.to_file(path)
        return path, save_file.cc

    @staticmethod
    def exit_editor(
        save_file: core.SaveFile | None = None, check_temp: bool = True
    ) -> NoReturn:
        """Exit the editor."""
        save_file_temp = None
        if check_temp:
            temp_path = core.SaveFile.get_temp_path()
            if temp_path.exists():
                try:
                    save_file_temp = core.SaveFile(temp_path.read())
                except core.SaveError as e:
                    tb = traceback.format_exc()
                    color.ColoredText.localize(
                        "save_temp_fail", error=str(e), traceback=tb
                    )
                    Main.leave()

        if save_file is None:
            save_file = save_file_temp
        if save_file is None:
            if check_temp:
                color.ColoredText.localize("save_temp_not_found")
            Main.leave()
        if save_file_temp is None:
            save_file_temp = save_file

        try:
            print()
            color.ColoredText.localize("checking_for_changes")
            if save_file.save_path is None:
                same = False
            else:
                same = save_file.save_path.read() == save_file.to_data()
        except core.SaveError:
            same = False

        if not same:
            color.ColoredText.localize("changes_found")
            print()
            save = color.ColoredInput().localize("save_before_exit") == "y"
            if save:
                save_management.SaveManagement.save_save(save_file)
        else:
            color.ColoredText.localize("no_changes")

        Main.leave()

    @staticmethod
    def leave() -> NoReturn:
        """Leave the editor."""
        color.ColoredText.localize("leave")
        sys.exit()



from __future__ import annotations
import base64
import time
from typing import Any

from bcsfe import core
import jwt

from bcsfe.cli import color


class RequestResult:
    def __init__(
        self,
        url: str,
        response: core.Response | None,
        headers: dict[str, str],
        data: str,
        payload: dict[str, Any] | None = None,
        timestamp: str | None = None,
    ):
        self.url = url
        self.response = response
        self.headers = headers
        self.data = data
        self.payload = payload
        self.timestamp = timestamp


class ServerHandler:
    auth_url = "https://nyanko-auth.ponosgames.com"
    save_url = "https://nyanko-save.ponosgames.com"
    backups_url = "https://nyanko-backups.ponosgames.com"
    aws_url = "https://nyanko-service-data-prd.s3.amazonaws.com"
    managed_item_url = "https://nyanko-managed-item.ponosgames.com"
    events_url = "https://nyanko-events.ponosgames.com"

    def __init__(self, save_file: core.SaveFile, print: bool = True):
        self.save_file = save_file
        self.print = print
        self.counter = 0

    @staticmethod
    def get_password_key() -> str:
        return "password"

    @staticmethod
    def get_auth_token_key() -> str:
        return "auth_token"

    @staticmethod
    def get_save_key_key() -> str:
        return "save_key"

    def save_password(self, password: str):
        self.save_file.store_string(ServerHandler.get_password_key(), password)

    def get_stored_password(self) -> str | None:
        return self.save_file.get_string(ServerHandler.get_password_key())

    def remove_stored_password(self):
        self.save_file.remove_string(ServerHandler.get_password_key())

    def save_save_key_data(self, save_key: dict[str, Any]):
        self.save_file.store_dict(ServerHandler.get_save_key_key(), save_key)

    def get_stored_save_key_data(self) -> dict[str, Any] | None:
        save_key_data = self.save_file.get_dict(ServerHandler.get_save_key_key())
        if save_key_data is None:
            return None
        if not self.validate_save_key_data(save_key_data):
            self.remove_stored_save_key_data()
            return None
        return save_key_data

    def validate_save_key_data(self, save_key_data: dict[str, Any]) -> bool:
        key = save_key_data.get("key")
        if key is None:
            return False
        if key.split("/")[2] != self.save_file.inquiry_code:
            return False
        policy = save_key_data.get("policy")
        if policy is None:
            return False
        policy = base64.b64decode(policy)
        json_policy = core.JsonFile.from_data(core.Data(policy)).to_object()
        expiration = json_policy.get("expiration")
        if expiration is None:
            return False
        expiration = int(
            time.mktime(time.strptime(expiration, "%Y-%m-%dT%H:%M:%S.%fZ"))
        )
        if expiration < time.time():
            return False
        return True

    def remove_stored_save_key_data(self):
        self.save_file.remove_dict(ServerHandler.get_save_key_key())

    def save_auth_token(self, auth_token: str):
        self.save_file.store_string(ServerHandler.get_auth_token_key(), auth_token)

    def get_stored_auth_token(self) -> str | None:
        token = self.save_file.get_string(ServerHandler.get_auth_token_key())
        return token

    def remove_stored_auth_token(self):
        self.save_file.remove_string(ServerHandler.get_auth_token_key())

    def get_password_new(self) -> str | None:
        self.print_key("getting_password")

        url = f"{self.auth_url}/v1/users"
        data = {
            "accountCode": self.save_file.inquiry_code,
            "accountCreatedAt": int(self.save_file.energy_penalty_timestamp),
            "nonce": core.Random.get_hex_string(32),
        }
        password = self.do_password_request(url, data)
        return password

    @staticmethod
    def log_error(key: str, result: RequestResult):
        if "EXPECT_THIS_TO_FAIL" in result.data:
            return
        if result.response is None:
            log_text = "Failed to make request. Check your internet connection."
            core.core_data.logger.log_error(log_text)
            return
        log_text = (
            f"Error: {key}\n"
            f"URL: {result.url}\n"
            f"Response Headers: {result.response.headers}\n"
            f"Response Body: {result.response.content.decode('utf-8')}\n"
            f"Status Code: {result.response.status_code}\n"
            f"Reason: {result.response.reason}\n"
            f"Request Headers: {result.headers}\n"
            f"Request Body: {result.data}\n"
        )
        core.core_data.logger.log_error(log_text)

    def do_password_request(self, url: str, dict_data: dict[str, Any]) -> str | None:
        result = self.do_request(url, dict_data)
        if result.payload is None:
            ServerHandler.log_error("password_fail", result)
            return None
        payload = result.payload
        password = payload.get("password", None)
        if password is None:
            ServerHandler.log_error("password_fail", result)
            self.remove_stored_password()
            return None
        password_refresh_token = payload.get("passwordRefreshToken", None)
        if password_refresh_token is None:
            ServerHandler.log_error("password_fail", result)
            self.remove_stored_password()
            return None
        account_code = payload.get("accountCode", None)
        timestamp = result.timestamp

        self.save_file.password_refresh_token = password_refresh_token
        self.save_password(password)
        if account_code:
            self.save_file.inquiry_code = account_code
            self.remove_stored_auth_token()
            self.remove_stored_save_key_data()

            if timestamp is not None:
                self.save_file.energy_penalty_timestamp = int(timestamp)
            if not self.update_managed_items():
                return None

        return password

    def do_request(self, url: str, dict_data: dict[str, Any]) -> RequestResult:
        data = (
            core.JsonFile.from_object(dict_data)
            .to_data(indent=None)
            .to_str()
            .replace(" ", "")
        )
        headers = core.AccountHeaders(self.save_file, data).get_headers()
        response = core.RequestHandler(url, headers, core.Data(data)).post()
        if response is None:
            self.log_no_internet(RequestResult(url, None, headers, data))
            return RequestResult(url, response, headers, data)
        json: dict[str, Any] = response.json()
        status_code = json.get("statusCode", 0)
        if status_code != 1:
            return RequestResult(url, response, headers, data)

        timestamp = json.get("timestamp", None)

        payload = json.get("payload", {})
        return RequestResult(url, response, headers, data, payload, timestamp)

    def refresh_password(self) -> str | None:
        self.print_key("refreshing_password")

        url = f"{self.auth_url}/v1/user/password"
        data = {
            "accountCode": self.save_file.inquiry_code,
            "passwordRefreshToken": self.save_file.password_refresh_token,
            "nonce": core.Random.get_hex_string(32),
        }
        return self.do_password_request(url, data)

    def get_auth_token_new(self, password: str) -> str | None:
        self.print_key("getting_auth_token")

        url = f"{self.auth_url}/v1/tokens"
        data = core.ClientInfo.from_save_file(self.save_file).get_client_info()
        data["password"] = password
        data["accountCode"] = self.save_file.inquiry_code

        result = self.do_request(url, data)
        if result.payload is None:
            ServerHandler.log_error("auth_token_fail", result)
            self.remove_stored_auth_token()
            self.remove_stored_password()
            return None
        payload = result.payload
        auth_token = payload.get("token", None)
        if auth_token is None:
            ServerHandler.log_error("auth_token_fail", result)
            self.remove_stored_auth_token()
            self.remove_stored_password()
            return None
        self.save_auth_token(auth_token)
        return auth_token

    def get_password(self, tries: int = 0) -> str | None:
        password = self.get_stored_password()
        if password is not None:
            return password
        password = self.refresh_password()
        if password is not None:
            return password
        password = self.get_password_new()
        if password is not None:
            return password
        self.create_new_account()
        if tries >= 1:
            return None
        return self.get_password(tries + 1)

    def validate_auth_token(self, auth_token: str) -> bool:
        token = jwt.decode(  # type: ignore
            auth_token,
            algorithms=["HS256"],
            options={"verify_signature": False},
        )
        if not token:
            return False
        if token.get("exp", 0) < time.time():
            return False
        if token.get("accountCode", None) != self.save_file.inquiry_code:
            return False

        return True

    def get_auth_token(self, tries: int = 1) -> str | None:
        auth_token = self.get_stored_auth_token()
        if auth_token is not None:
            if self.validate_auth_token(auth_token):
                return auth_token
            self.remove_stored_auth_token()
        password = self.get_password()
        if password is None:
            return None
        auth_token = self.get_stored_auth_token()
        if auth_token is not None:
            return auth_token
        auth_token = self.get_auth_token_new(password)
        if auth_token is not None:
            return auth_token

        if tries > 0:
            self.print_key("retry_auth_token")
            return self.get_auth_token(tries - 1)

        return None

    def log_no_internet(self, result: RequestResult):
        ServerHandler.log_error("no_internet", result)
        if self.print:
            core.print_no_internet()

    def get_save_key_new(self, auth_token: str) -> dict[str, Any] | None:
        self.print_key("getting_save_key")

        nonce = core.Random.get_hex_string(32)
        url = f"{self.save_url}/v2/save/key?nonce={nonce}"
        headers = {
            "accept-encoding": "gzip",
            "connection": "keep-alive",
            "authorization": "Bearer " + auth_token,
            "nyanko-timestamp": str(int(time.time())),
            "user-agent": "Dalvik/2.1.0 (Linux; U; Android 9; SM-G955F Build/N2G48B)",
        }
        response = core.RequestHandler(url, headers).get()
        if response is None:
            self.log_no_internet(RequestResult(url, None, headers, ""))
            return None
        json: dict[str, Any] = response.json()
        status_code = json.get("statusCode", 0)
        if status_code != 1:
            ServerHandler.log_error(
                "save_key_fail", RequestResult(url, response, headers, "")
            )
            self.remove_stored_auth_token()
            return None
        payload = json.get("payload", {})
        self.save_save_key_data(payload)
        return payload

    def get_save_key(self) -> dict[str, Any] | None:
        # save_key = self.get_stored_save_key_data()
        # if save_key and save_key.get("key", None):
        #    return save_key
        auth_token = self.get_auth_token()
        if auth_token is None:
            return None
        # save_key = self.get_stored_save_key_data()
        # if save_key:
        #    return save_key
        save_key = self.get_save_key_new(auth_token)
        if save_key is not None:
            return save_key

        return None

    def get_upload_request_form(
        self,
        save_key: dict[str, str],
    ) -> core.MultipartForm:
        save_data = self.save_file.to_data()
        form_data = core.MultipartForm()
        for key, value in save_key.items():
            if key == "url":
                continue
            form_data.add_key(key, value.encode(), "text/plain")

        form_data.add_key(
            "file", save_data.to_bytes(), "application/octet-stream", "file.sav"
        )
        return form_data

    def upload_save_data(self, save_key: dict[str, Any]) -> bool:
        self.print_key("uploading_save_file")

        form = self.get_upload_request_form(save_key)
        if form is None:
            self.remove_stored_save_key_data()
            return False
        url = save_key.get("url")
        if url is None:
            url = f"{self.aws_url}/"
        headers = {
            "accept-encoding": "gzip",
            "connection": "keep-alive",
            "user-agent": "Dalvik/2.1.0 (Linux; U; Android 9; SM-G955F Build/N2G48B)",
        }
        response = core.RequestHandler(url, headers, form=form).post(no_timeout=True)
        if response is None:
            self.log_no_internet(RequestResult(url, None, headers, ""))
            return False
        if response.status_code != 204:
            ServerHandler.log_error(
                "upload_fail_aws",
                RequestResult(
                    url,
                    response,
                    headers,
                    form.get_all_type("text-plain"),
                ),
            )

            self.remove_stored_save_key_data()
            return False
        return True

    def print_key(self, key: str, **kwargs: Any):
        if self.print:
            color.ColoredText.localize(key, **kwargs)

    def get_codes(self, upload_managed_items: bool = True) -> tuple[str, str] | None:
        self.save_file.show_ban_message = False

        auth_token = self.get_auth_token()
        if auth_token is None:
            return None

        save_key = self.get_save_key()

        if save_key is None:
            self.remove_stored_save_key_data()
            return None

        if not self.upload_save_data(save_key):
            return None

        self.print_key("getting_codes")

        bmd = core.BackupMetaData(self.save_file)
        meta_data = bmd.create(save_key["key"], upload_managed_items)

        url = f"{self.save_url}/v2/transfers"
        headers = core.AccountHeaders(self.save_file, meta_data).get_headers()
        headers["authorization"] = "Bearer " + auth_token

        response = core.RequestHandler(url, headers, core.Data(meta_data)).post()
        if response is None:
            self.log_no_internet(RequestResult(url, None, headers, meta_data))
            return None
        json: dict[str, Any] = response.json()
        status_code = json.get("statusCode", 0)
        if status_code != 1:
            ServerHandler.log_error(
                "upload_fail_transfers",
                RequestResult(url, response, headers, meta_data),
            )
            self.remove_stored_auth_token()
            return None
        payload = json.get("payload", {})
        transfer_code = payload.get("transferCode", None)
        confirmation_code = payload.get("pin", None)
        if transfer_code is None or confirmation_code is None:
            ServerHandler.log_error(
                "upload_fail_transfers",
                RequestResult(url, response, headers, ""),
            )
            self.remove_stored_auth_token()
            return None
        bmd.remove_managed_items()
        if self.print:
            print()
        return (transfer_code, confirmation_code)

    def has_managed_items(self) -> bool:
        bmd = core.BackupMetaData(self.save_file)
        managed_items = bmd.get_managed_items()
        if len(managed_items) == 0:
            return False
        return True

    def upload_meta_data(self) -> bool:
        auth_token = self.get_auth_token()
        if auth_token is None:
            return False

        save_key = self.get_save_key()
        if save_key is None:
            self.remove_stored_save_key_data()
            return False

        if not self.upload_save_data(save_key):
            return False

        bmd = core.BackupMetaData(self.save_file)
        meta_data = bmd.create(save_key["key"])

        url = f"{self.save_url}/v2/backups"
        headers = core.AccountHeaders(self.save_file, meta_data).get_headers()
        headers["authorization"] = "Bearer " + auth_token

        response = core.RequestHandler(url, headers, core.Data(meta_data)).post()
        if response is None:
            self.log_no_internet(RequestResult(url, None, headers, meta_data))
            return False
        json: dict[str, Any] = response.json()
        status_code = json.get("statusCode", 0)
        if status_code != 1:
            self.remove_stored_auth_token()
            return False
        bmd.remove_managed_items()
        return True

    def get_new_inquiry_code(self) -> str | None:
        url = f"{self.backups_url}/?action=createAccount&referenceId="

        response = core.RequestHandler(url).get()
        if response is None:
            self.log_no_internet(RequestResult(url, None, {}, ""))
            return None
        data = response.json()
        iq = data["accountId"]
        return iq

    def create_new_account(self) -> bool:
        new_iq = self.get_new_inquiry_code()
        if new_iq is None:
            return False
        self.save_file.inquiry_code = new_iq
        self.remove_stored_auth_token()
        self.remove_stored_save_key_data()
        self.remove_stored_password()
        fail_text = "EXPECT_THIS_TO_FAIL"
        start_count = (40 - len(fail_text)) // 2
        end_count = 40 - len(fail_text) - start_count
        self.save_file.password_refresh_token = (
            "_" * start_count + fail_text + "_" * end_count
        )
        password = self.get_password()
        auth_token = self.get_auth_token()
        save_key_data = self.get_save_key()
        self.update_managed_items()
        self.save_file.show_ban_message = False
        if password is None or auth_token is None or save_key_data is None:
            return False
        return True

    @staticmethod
    def from_codes(
        transfer_code: str,
        confirmation_code: str,
        cc: core.CountryCode,
        gv: core.GameVersion,
        print: bool = True,
        save_backup: bool = True,
    ) -> tuple[ServerHandler | None, RequestResult | None]:
        url = f"{ServerHandler.save_url}/v2/transfers/{transfer_code}/reception"
        data = core.ClientInfo(cc, gv).get_client_info()
        data["pin"] = confirmation_code
        data_str = (
            core.JsonFile.from_object(data)
            .to_data(indent=None)
            .to_str()
            .replace(" ", "")
        )

        headers = {
            "content-type": "application/json",
            "accept-encoding": "gzip",
            "connection": "keep-alive",
            "user-agent": "Dalvik/2.1.0 (Linux; U; Android 9; SM-G955F Build/N2G48B)",
        }
        response = core.RequestHandler(url, headers, core.Data(data_str)).post()
        if response is None:
            if print:
                core.print_no_internet()
            return None, None
        resp_headers = response.headers
        content_type = resp_headers.get("content-type", "")
        if content_type != "application/octet-stream":
            return None, RequestResult(url, response, headers, data_str)

        save_data = response.content

        if save_backup:
            temp_path = (
                core.Path.get_documents_folder()
                .add("saves")
                .generate_dirs()
                .add("transfer_backup")
            )
            try:
                temp_path.write(core.Data(save_data))
            except Exception as e:
                color.ColoredText.localize(
                    "transfer_backup_fail", path=str(temp_path), error=e
                )
            else:
                if print:
                    color.ColoredText.localize("transfer_backup", path=str(temp_path))

        save_file = core.SaveFile(core.Data(save_data), cc=cc)

        password_refresh_token = resp_headers.get("Nyanko-Password-Refresh-Token")
        if password_refresh_token is not None:
            save_file.password_refresh_token = password_refresh_token

        server_handler = ServerHandler(save_file)
        password = resp_headers.get("Nyanko-Password")
        if password is not None:
            server_handler.save_password(password)

        return server_handler, RequestResult(url, response, headers, data_str)

    def update_managed_items(self) -> bool:
        auth_token = self.get_auth_token()
        if auth_token is None:
            return False
        data = {
            "catfoodAmount": self.save_file.catfood,
            "isPaid": True,
            "legendTicketAmount": self.save_file.legend_tickets,
            "nonce": core.Random.get_hex_string(32),
            "platinumTicketAmount": self.save_file.platinum_tickets,
            "rareTicketAmount": self.save_file.rare_tickets,
        }
        data_str = (
            core.JsonFile.from_object(data)
            .to_data(indent=None)
            .to_str()
            .replace(" ", "")
        )
        url = f"{self.managed_item_url}/v1/managed-items"
        headers = core.AccountHeaders(self.save_file, data_str).get_headers()
        headers["authorization"] = "Bearer " + auth_token
        response = core.RequestHandler(url, headers, core.Data(data_str)).post()
        if response is None:
            self.log_no_internet(RequestResult(url, None, headers, data_str))
            return False
        json: dict[str, Any] = response.json()
        status_code = json.get("statusCode", 0)
        if status_code != 1:
            self.remove_stored_auth_token()
            return False

        core.BackupMetaData(self.save_file).remove_managed_items()
        return True

    def download_event_data(self, filename: str) -> core.Data | None:
        url = (
            self.events_url
            + f"/battlecats{self.save_file.cc.get_patching_code()}_production/{filename}"
        )

        auth_token = self.get_auth_token()

        if auth_token is None:
            return None

        url += f"?jwt={auth_token}"

        headers = {
            "accept-encoding": "gzip",
            "connection": "keep-alive",
            "user-agent": "Dalvik/2.1.0 (Linux; U; Android 9; Pixel 2 Build/PQ3A.190801.002)",
        }

        resp = core.RequestHandler(url, headers).get()

        if resp is None:
            return None

        return core.Data(resp.content)

    def download_gatya_data(self) -> core.Data | None:
        return self.download_event_data("gatya.tsv")

    def download_item_data(self) -> core.Data | None:
        return self.download_event_data("item.tsv")

    def download_sale_data(self) -> core.Data | None:
        return self.download_event_data("sale.tsv")



from __future__ import annotations

from collections.abc import Callable
from bcsfe import core


def max_catfood(save_file: core.SaveFile):
    orig = save_file.catfood
    save_file.catfood = core.core_data.max_value_manager.get(core.MaxValueType.CATFOOD)
    core.BackupMetaData(save_file).add_managed_item(
        core.ManagedItem.from_change(
            save_file.catfood - orig, core.ManagedItemType.CATFOOD
        )
    )


def max_rare_tickets(save_file: core.SaveFile):
    orig = save_file.rare_tickets
    save_file.rare_tickets = core.core_data.max_value_manager.get(
        core.MaxValueType.RARE_TICKETS
    )
    core.BackupMetaData(save_file).add_managed_item(
        core.ManagedItem.from_change(
            save_file.rare_tickets - orig, core.ManagedItemType.RARE_TICKET
        )
    )


def max_plat_tickets(save_file: core.SaveFile):
    orig = save_file.platinum_tickets
    save_file.platinum_tickets = core.core_data.max_value_manager.get(
        core.MaxValueType.PLATINUM_TICKETS
    )
    core.BackupMetaData(save_file).add_managed_item(
        core.ManagedItem.from_change(
            save_file.platinum_tickets - orig, core.ManagedItemType.PLATINUM_TICKET
        )
    )


def max_plat_shards(save_file: core.SaveFile):
    save_file.platinum_shards = 10 * core.core_data.max_value_manager.get(
        core.MaxValueType.PLATINUM_TICKETS
    )


def max_legend_tickets(save_file: core.SaveFile):
    orig = save_file.legend_tickets
    save_file.legend_tickets = core.core_data.max_value_manager.get(
        core.MaxValueType.LEGEND_TICKETS
    )
    core.BackupMetaData(save_file).add_managed_item(
        core.ManagedItem.from_change(
            save_file.legend_tickets - orig, core.ManagedItemType.LEGEND_TICKET
        )
    )


def max_xp(save_file: core.SaveFile):
    save_file.xp = core.core_data.max_value_manager.get(core.MaxValueType.XP)


def max_np(save_file: core.SaveFile):
    save_file.np = core.core_data.max_value_manager.get(core.MaxValueType.NP)


def max_100_million_ticket(save_file: core.SaveFile):
    save_file.hundred_million_ticket = core.core_data.max_value_manager.get(
        core.MaxValueType.HUNDRED_MILLION_TICKETS
    )


def max_leadership(save_file: core.SaveFile):
    save_file.leadership = core.core_data.max_value_manager.get(
        core.MaxValueType.LEADERSHIP
    )


def max_battle_items(save_file: core.SaveFile):
    for item in save_file.battle_items.items:
        item.amount = core.core_data.max_value_manager.get(
            core.MaxValueType.BATTLE_ITEMS
        )


def max_catseyes(save_file: core.SaveFile):
    for id in range(len(save_file.catseyes)):
        save_file.catseyes[id] = core.core_data.max_value_manager.get(
            core.MaxValueType.CATSEYES
        )


def max_treasure_chests(save_file: core.SaveFile):
    for id in range(len(save_file.treasure_chests)):
        save_file.treasure_chests[id] = core.core_data.max_value_manager.get(
            core.MaxValueType.TREASURE_CHESTS
        )


def max_catamins(save_file: core.SaveFile):
    for id in range(len(save_file.catseyes)):
        save_file.catamins[id] = core.core_data.max_value_manager.get(
            core.MaxValueType.CATAMINS
        )


def max_labyrinth_medals(save_file: core.SaveFile):
    for id in range(len(save_file.labyrinth_medals)):
        save_file.labyrinth_medals[id] = core.core_data.max_value_manager.get(
            core.MaxValueType.LABYRINTH_MEDALS
        )


# def max_catfruit(save_file: core.SaveFile):
#     for id in range(len(save_file.catfruit)):
#         save_file.catfruit[id] = core.core_data.max_value_manager.get_new(
#             core.MaxValueType.CATFRUIT
#         )


def max_normal_tickets(save_file: core.SaveFile):
    save_file.normal_tickets = core.core_data.max_value_manager.get(
        core.MaxValueType.NORMAL_TICKETS
    )


def max_all(save_file: core.SaveFile):
    maxes = core.core_data.max_value_manager
    features: dict[str, Callable[[core.SaveFile], None]] = {
        "catfood": max_catfood,
        "xp": max_xp,
        "normal_tickets": max_normal_tickets,
        "rare_tickets": max_rare_tickets,
        "platinum_tickets": max_plat_tickets,
        "legend_tickets": max_legend_tickets,
        "platinum_shards": max_plat_shards,
        "np": max_np,
        "leadership": max_leadership,
        "battle_items": max_battle_items,
        "catseyes": max_catseyes,
        "catamins": max_catamins,
        "labyrinth_medals": max_labyrinth_medals,
        "100_million_ticket": max_100_million_ticket,
        "treasure_chests": max_treasure_chests,
    }
    # TODO: finish



from __future__ import annotations
from bcsfe import core
from bcsfe.cli import color


def clear_tutorial(
    save_file: core.SaveFile, display_already_cleared: bool = True
):
    core.StoryChapters.clear_tutorial(save_file)
    if display_already_cleared:
        color.ColoredText.localize("tutorial_cleared")



from __future__ import annotations
import enum
from typing import Any, Callable
from bcsfe import core
from bcsfe.cli import dialog_creator, color


class Gatya:
    def __init__(self, rare_seed: int, normal_seed: int):
        self.rare_seed = rare_seed
        self.normal_seed = normal_seed
        self.event_seed = 0
        self.stepup_stage_3_cooldown = 0
        self.previous_normal_roll = 0
        self.previous_normal_roll_type = 0
        self.previous_rare_roll = 0
        self.previous_rare_roll_type = 0
        self.unknown1 = False
        self.roll_single = False
        self.roll_multi = False
        self.trade_progress = 0
        self.step_up_stages: dict[int, int] = {}
        self.stepup_durations: dict[int, float] = {}

        self.gatya_data_set: GatyaDataSet | None = None

    @staticmethod
    def init() -> Gatya:
        return Gatya(0, 0)

    @staticmethod
    def read_rare_normal_seed(data: core.Data, gv: core.GameVersion) -> Gatya:
        if gv < 33:
            return Gatya(data.read_ulong(), data.read_ulong())
        return Gatya(data.read_uint(), data.read_uint())

    def read_event_seed(self, data: core.Data, gv: core.GameVersion):
        if gv < 33:
            self.event_seed = data.read_ulong()
        else:
            self.event_seed = data.read_uint()

    def write_rare_normal_seed(self, data: core.Data):
        data.write_uint(self.rare_seed)
        data.write_uint(self.normal_seed)

    def write_event_seed(self, data: core.Data):
        data.write_uint(self.event_seed)

    def read2(self, data: core.Data):
        self.stepup_stage_3_cooldown = data.read_int()
        self.previous_normal_roll = data.read_int()
        self.previous_normal_roll_type = data.read_int()
        self.previous_rare_roll = data.read_int()
        self.previous_rare_roll_type = data.read_int()
        self.unknown1 = data.read_bool()
        self.roll_single = data.read_bool()
        self.roll_multi = data.read_bool()

    def write2(self, data: core.Data):
        data.write_int(self.stepup_stage_3_cooldown)
        data.write_int(self.previous_normal_roll)
        data.write_int(self.previous_normal_roll_type)
        data.write_int(self.previous_rare_roll)
        data.write_int(self.previous_rare_roll_type)
        data.write_bool(self.unknown1)
        data.write_bool(self.roll_single)
        data.write_bool(self.roll_multi)

    def read_trade_progress(self, data: core.Data):
        self.trade_progress = data.read_int()

    def write_trade_progress(self, data: core.Data):
        data.write_int(self.trade_progress)

    def read_stepup(self, data: core.Data):
        self.step_up_stages: dict[int, int] = {}
        total = data.read_int()
        for _ in range(total):
            key = data.read_int()
            self.step_up_stages[key] = data.read_int()

        self.stepup_durations: dict[int, float] = {}
        total = data.read_int()
        for _ in range(total):
            key = data.read_int()
            self.stepup_durations[key] = data.read_double()

    def write_stepup(self, data: core.Data):
        data.write_int(len(self.step_up_stages))
        for id, stage in self.step_up_stages.items():
            data.write_int(id)
            data.write_int(stage)

        data.write_int(len(self.stepup_durations))
        for id, duration in self.stepup_durations.items():
            data.write_int(id)
            data.write_double(duration)

    def serialize(self) -> dict[str, Any]:
        return {
            "rare_seed": self.rare_seed,
            "normal_seed": self.normal_seed,
            "stepup_stage_3_cooldown": self.stepup_stage_3_cooldown,
            "previous_normal_roll": self.previous_normal_roll,
            "previous_normal_roll_type": self.previous_normal_roll_type,
            "previous_rare_roll": self.previous_rare_roll,
            "previous_rare_roll_type": self.previous_rare_roll_type,
            "unknown1": self.unknown1,
            "roll_single": self.roll_single,
            "roll_multi": self.roll_multi,
            "trade_progress": self.trade_progress,
            "event_seed": self.event_seed,
            "step_up_stages": self.step_up_stages,
            "stepup_durations": self.stepup_durations,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Gatya:
        gatya = Gatya(data.get("rare_seed", 0), data.get("normal_seed", 0))
        gatya.stepup_stage_3_cooldown = data.get("stepup_stage_3_cooldown", 0)
        gatya.previous_normal_roll = data.get("previous_normal_roll", 0)
        gatya.previous_normal_roll_type = data.get("previous_normal_roll_type", 0)
        gatya.previous_rare_roll = data.get("previous_rare_roll", 0)
        gatya.previous_rare_roll_type = data.get("previous_rare_roll_type", 0)
        gatya.unknown1 = data.get("unknown1", False)
        gatya.roll_single = data.get("roll_single", False)
        gatya.roll_multi = data.get("roll_multi", False)
        gatya.trade_progress = data.get("trade_progress", 0)
        gatya.event_seed = data.get("event_seed", 0)
        gatya.step_up_stages = data.get("step_up_stages", {})
        gatya.stepup_durations = data.get("stepup_durations", {})
        return gatya

    def __repr__(self) -> str:
        return f"Gatya({self.serialize()})"

    def __str__(self) -> str:
        return f"Gatya({self.serialize()})"

    def edit_rare_gatya_seed(self):
        self.rare_seed = dialog_creator.SingleEditor(
            "rare_gatya_seed",
            self.rare_seed,
            None,
            localized_item=True,
            signed=False,
        ).edit()

    def edit_normal_gatya_seed(self):
        self.normal_seed = dialog_creator.SingleEditor(
            "normal_gatya_seed",
            self.normal_seed,
            None,
            localized_item=True,
            signed=False,
        ).edit()

    def edit_event_gatya_seed(self):
        self.event_seed = dialog_creator.SingleEditor(
            "event_gatya_seed",
            self.event_seed,
            None,
            localized_item=True,
            signed=False,
        ).edit()

    def read_gatya_data_set(self, save_file: core.SaveFile) -> GatyaDataSet:
        if self.gatya_data_set is not None:
            return self.gatya_data_set
        self.gatya_data_set = GatyaDataSet(save_file)
        return self.gatya_data_set


class GatyaDataSet:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.gatya_data_set = self.load_gatya_data_set("R", 1)

    def load_gatya_data_set(self, rarity: str, id: int) -> list[list[int]] | None:
        file_name = f"GatyaDataSet{rarity.upper()[0]}{id}.csv"
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", file_name)
        if data is None:
            return None
        csv = core.CSV(data)
        dt: list[list[int]] = []
        for line in csv:
            cat_ids: list[int] = []
            for cat_id in line:
                cat_id = cat_id.to_int()
                if cat_id != -1:
                    cat_ids.append(cat_id)
            dt.append(cat_ids)
        return dt

    def get_cat_ids(self, gatya_id: int) -> list[int] | None:
        if self.gatya_data_set is None:
            return None
        try:
            return self.gatya_data_set[gatya_id]
        except IndexError:
            return None


class GatyaInfo:
    def __init__(self, gatya_id: int, cc: core.CountryCode, type_str: str = "R"):
        self.gatya_id = gatya_id
        self.cc = cc
        self.gatya_data_set: GatyaDataSet | None = None
        self.type = type_str
        self.data: core.Data | None = None

    def get_id_str(self) -> str:
        return f"{self.gatya_id:03}"

    def get_cc_str(self) -> str:
        if self.cc == core.CountryCode("jp"):
            return ""
        return self.cc.get_patching_code() + "/"

    def get_url(self) -> str:
        return f"https://ponosgames.com/information/appli/battlecats/gacha/rare/{self.get_cc_str()}{self.type}{self.get_id_str()}.html"

    def download_data(self) -> core.Data | None:
        url = self.get_url()

        response = core.RequestHandler(url).get()
        if response is None:
            return
        data = core.Data(response.content)

        self.save_data(data)
        return data

    def get_file_path(self) -> core.Path:
        return (
            core.Path.get_documents_folder()
            .add("other_game_data")
            .add(self.cc.get_code())
            .add("gatya_info")
            .generate_dirs()
            .add(f"{self.type}{self.get_id_str()}.html")
        )

    def save_data(self, data: core.Data):
        try:
            data.to_file(self.get_file_path())
        except Exception as e:
            color.ColoredText.localize("save_gatya_error", error=e)
        self.data = data

    def load_data_from_file(self) -> core.Data | None:
        if not self.get_file_path().exists():
            return None
        return core.Data.from_file(self.get_file_path())

    def get_data(self) -> core.Data | None:
        if self.data is not None:
            return self.data
        data = self.load_data_from_file()
        if data is None:
            data = self.download_data()
        return data

    def get_name(self) -> str | None:
        data = self.get_data()
        if data is None:
            return None
        # find <h2>...</h2>
        data = data.get_bytes()
        h2 = data.find(b"<h2>")
        if h2 == -1:
            return None
        h2_end = data.find(b"</h2>", h2)
        if h2_end == -1:
            return None
        text = data[h2 + 4 : h2_end].decode("utf-8")
        # remove <span...</span>
        span = text.find("<span")
        if span == -1:
            return text
        span_end = text.find("</span>", span)
        if span_end == -1:
            return text
        return text[:span] + text[span_end + 7 :]


class GatyaInfos:
    def __init__(self, save_file: core.SaveFile, type_str: str = "R", set_id: int = 1):
        self.save_file = save_file
        self.type = type_str
        self.set_id = set_id
        self.gatya_data_set = GatyaDataSet(save_file).load_gatya_data_set(
            type_str, set_id
        )
        self.infos: list[GatyaInfo] = []
        self.got_all = False

    def get_all(
        self,
        threaded: bool = True,
        print_progress: bool = True,
        max_threads: int = 16,
    ):
        if self.gatya_data_set is None:
            return
        all_ids = len(self.gatya_data_set)
        if threaded:
            funcs: list[Callable[..., Any]] = []
            args: list[list[Any]] = []
            for id in range(all_ids):
                funcs.append(self.get)
                args.append([id, print_progress])
            core.thread_run_many(funcs, args, max_threads=max_threads)

        else:
            for id in range(all_ids):
                self.infos.append(self.get(id, print_progress=print_progress))

        self.got_all = True

    def get(self, gatya_id: int, print_progress: bool):
        if print_progress:
            color.ColoredText.localize(
                "gatya_info_progress",
                current=len(self.infos or []) + 1,
                total=len(self.gatya_data_set or []),
            )
        info = GatyaInfo(gatya_id, self.save_file.cc, self.type)
        info.get_data()
        self.infos.append(info)
        return info

    def get_info(self, gatya_id: int) -> GatyaInfo | None:
        if self.infos:
            return self.infos[gatya_id]
        return None

    def get_all_names(self) -> dict[int, str]:
        if not self.got_all:
            self.get_all(True, max_threads=64)
        names: dict[int, str] = {}
        for info in self.infos:
            names[
                info.gatya_id
            ] = info.get_name() or core.core_data.local_manager.get_key(
                "unknown_banner"
            )

        return names


class GatyaDataOptionSet:
    def __init__(
        self,
        id: int,
        banner_on: bool,
        ticket_item_id: int,
        anim_id: int,
        button_cut_id: int,
        series_id: int,
        menu_cut_id: int,
        char_id: int | None,
        wait_maanim: bool | None,
    ):
        self.id = id
        self.banner_on = banner_on
        self.ticket_item_id = ticket_item_id
        self.anim_id = anim_id
        self.button_cut_id = button_cut_id
        self.series_id = series_id
        self.menu_cut_id = menu_cut_id
        self.char_id = char_id
        self.wait_maanim = wait_maanim

    @staticmethod
    def from_csv_row(row: core.Row) -> GatyaDataOptionSet:
        return GatyaDataOptionSet(
            row.next_int(),
            row.next_bool(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int(),
            row.next_int_opt(),
            row.next_bool_opt(),
        )


class GatyaEventType(enum.Enum):
    NORMAL = "N"
    RARE = "R"
    EVENT = "E"


class GatyaDataOption:
    def __init__(self, sets: list[GatyaDataOptionSet]):
        self.sets = sets

    def get(self, set_id: int) -> GatyaDataOptionSet | None:
        for gset in self.sets:
            if gset.id == set_id:
                return gset

        return None

    @staticmethod
    def from_csv(csv: core.CSV) -> GatyaDataOption:
        sets: list[GatyaDataOptionSet] = []
        csv.read_line()  # skip headers
        for row in csv:
            sets.append(GatyaDataOptionSet.from_csv_row(row))

        return GatyaDataOption(sets)

    @staticmethod
    def from_data(data: core.Data) -> GatyaDataOption:
        return GatyaDataOption.from_csv(core.CSV(data, "\t"))

    @staticmethod
    def get_filename(event_type: GatyaEventType) -> str:
        return f"GatyaData_Option_Set{event_type.value}.tsv"

    @staticmethod
    def read(
        save_file: core.SaveFile, e_type: GatyaEventType
    ) -> GatyaDataOption | None:
        gdg = core.core_data.get_game_data_getter(save_file)

        data = gdg.download("DataLocal", GatyaDataOption.get_filename(e_type))
        if data is None:
            return None

        return GatyaDataOption.from_data(data)



from bcsfe.cli.edits import (
    basic_items,
    cat_editor,
    clear_tutorial,
    rare_ticket_trade,
    fixes,
    enemy_editor,
    aku_realm,
    map,
    event_tickets,
    max_all,
    storage,
)

__all__ = [
    "basic_items",
    "cat_editor",
    "clear_tutorial",
    "rare_ticket_trade",
    "fixes",
    "enemy_editor",
    "aku_realm",
    "map",
    "event_tickets",
    "max_all",
    "storage",
]



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import color, dialog_creator


class Medals:
    def __init__(
        self,
        u1: int,
        u2: int,
        u3: int,
        medal_data_1: list[int],
        medal_data_2: dict[int, int],
        ub: bool,
    ):
        self.u1 = u1
        self.u2 = u2
        self.u3 = u3
        self.medal_data_1 = medal_data_1
        self.medal_data_2 = medal_data_2
        self.ub = ub

    @staticmethod
    def init() -> Medals:
        return Medals(0, 0, 0, [], {}, False)

    @staticmethod
    def read(data: core.Data) -> Medals:
        u1 = data.read_int()
        u2 = data.read_int()
        u3 = data.read_int()
        total_medals = data.read_short()
        medal_data_1 = data.read_short_list(total_medals)
        total_medals = data.read_short()
        medal_data_2: dict[int, int] = {}
        for _ in range(total_medals):
            key = data.read_short()
            value = data.read_byte()
            medal_data_2[key] = value
        ub = data.read_bool()
        return Medals(u1, u2, u3, medal_data_1, medal_data_2, ub)

    def write(self, data: core.Data) -> None:
        data.write_int(self.u1)
        data.write_int(self.u2)
        data.write_int(self.u3)
        data.write_short(len(self.medal_data_1))
        data.write_short_list(self.medal_data_1, write_length=False)
        data.write_short(len(self.medal_data_2))
        for key, value in self.medal_data_2.items():
            data.write_short(key)
            data.write_byte(value)
        data.write_bool(self.ub)

    def serialize(self) -> dict[str, Any]:
        return {
            "u1": self.u1,
            "u2": self.u2,
            "u3": self.u3,
            "medal_data_1": self.medal_data_1,
            "medal_data_2": self.medal_data_2,
            "ub": self.ub,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Medals:
        return Medals(
            data.get("u1", 0),
            data.get("u2", 0),
            data.get("u3", 0),
            data.get("medal_data_1", []),
            data.get("medal_data_2", {}),
            data.get("ub", False),
        )

    def __repr__(self) -> str:
        return (
            f"Medals(u1={self.u1}, u2={self.u2}, u3={self.u3}, "
            f"medal_data_1={self.medal_data_1}, medal_data_2={self.medal_data_2}, "
            f"ub={self.ub})"
        )

    def __str__(self) -> str:
        return self.__repr__()

    def has_medal(self, medal_id: int) -> bool:
        return medal_id in self.medal_data_1

    @staticmethod
    def edit_medals(save_file: core.SaveFile):
        medals = save_file.medals
        medal_names = core.core_data.get_medal_names(save_file)
        if medal_names.medal_names is None:
            return
        options = ["add_medals", "remove_medals"]
        choice = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="medal_add_remove_dialog", single_choice=True
        ).single_choice()
        if choice is None:
            return
        choice -= 1
        add_medals = choice == 0

        medals_to_choose_from: list[tuple[int, str]] = []
        for i, medal in enumerate(medal_names.medal_names):
            if len(medal) == 0:
                continue
            if medals.has_medal(i) == add_medals:
                continue
            key = "medal_string"
            string = core.core_data.local_manager.get_key(
                key, medal_name=medal[0], medal_req=medal[1]
            )
            medals_to_choose_from.append((i, string))
        if len(medals_to_choose_from) == 0:
            return
        options = [medal[1] for medal in medals_to_choose_from]
        choices, _ = dialog_creator.ChoiceInput.from_reduced(
            options, dialog="select_medals"
        ).multiple_choice()
        if choices is None:
            return
        for choice in choices:
            medal_id = medals_to_choose_from[choice][0]
            if add_medals:
                medals.add_medal(medal_id)
            else:
                medals.remove_medal(medal_id)

        if add_medals:
            color.ColoredText.localize("medals_added")
        else:
            color.ColoredText.localize("medals_removed")

    def add_medal(self, medal_id: int) -> None:
        if self.has_medal(medal_id):
            return
        self.medal_data_1.append(medal_id)
        self.medal_data_2[medal_id] = 0

    def remove_medal(self, medal_id: int) -> None:
        if medal_id in self.medal_data_2:
            del self.medal_data_2[medal_id]
        if medal_id in self.medal_data_1:
            self.medal_data_1.remove(medal_id)


class MedalNames:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.medal_names = self.get_medal_names()

    def get_medal_names(self) -> list[list[str]] | None:
        file_name = "medalname.tsv"
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("resLocal", file_name)
        if data is None:
            return None
        csv = core.CSV(data, delimiter="\t")
        names: list[list[str]] = []
        for row in csv:
            names.append(row.to_str_list())
        return names

    def get_medal_name(self, medal_id: int) -> list[str] | None:
        if self.medal_names is None:
            return None
        if medal_id < 0 or medal_id >= len(self.medal_names):
            return []
        return self.medal_names[medal_id]



from __future__ import annotations
from typing import Any
from bcsfe import core


class Login:
    def __init__(self, count: int):
        self.count = count

    @staticmethod
    def init() -> Login:
        return Login(0)

    @staticmethod
    def read(stream: core.Data) -> Login:
        count = stream.read_int()
        return Login(count)

    def write(self, stream: core.Data):
        stream.write_int(self.count)

    def serialize(self) -> int:
        return self.count

    @staticmethod
    def deserialize(data: int) -> Login:
        return Login(data)

    def __repr__(self):
        return f"Login({self.count})"

    def __str__(self):
        return f"Login({self.count})"


class Logins:
    def __init__(self, logins: list[Login]):
        self.logins = logins

    @staticmethod
    def init() -> Logins:
        return Logins([])

    @staticmethod
    def read(stream: core.Data) -> Logins:
        total = stream.read_int()
        logins: list[Login] = []
        for _ in range(total):
            logins.append(Login.read(stream))
        return Logins(logins)

    def write(self, stream: core.Data):
        stream.write_int(len(self.logins))
        for login in self.logins:
            login.write(stream)

    def serialize(self) -> list[int]:
        return [login.serialize() for login in self.logins]

    @staticmethod
    def deserialize(data: list[int]) -> Logins:
        return Logins([Login.deserialize(login) for login in data])

    def __repr__(self):
        return f"Logins({self.logins})"

    def __str__(self):
        return f"Logins({self.logins})"


class LoginSets:
    def __init__(self, logins: list[Logins]):
        self.logins = logins

    @staticmethod
    def init() -> LoginSets:
        return LoginSets([])

    @staticmethod
    def read(stream: core.Data) -> LoginSets:
        total = stream.read_int()
        logins: list[Logins] = []
        for _ in range(total):
            logins.append(Logins.read(stream))
        return LoginSets(logins)

    def write(self, stream: core.Data):
        stream.write_int(len(self.logins))
        for login in self.logins:
            login.write(stream)

    def serialize(self) -> list[list[int]]:
        return [login.serialize() for login in self.logins]

    @staticmethod
    def deserialize(data: list[list[int]]) -> LoginSets:
        return LoginSets([Logins.deserialize(login) for login in data])

    def __repr__(self):
        return f"LoginSets({self.logins})"

    def __str__(self):
        return f"LoginSets({self.logins})"


class LoginBonus:
    def __init__(
        self,
        old_logins: LoginSets | None = None,
        logins: dict[int, Login] | None = None,
    ):
        self.old_logins = old_logins
        self.logins = logins

    @staticmethod
    def init(gv: core.GameVersion) -> LoginBonus:
        if gv < 80000:
            return LoginBonus(old_logins=LoginSets.init())
        else:
            return LoginBonus(logins={})

    @staticmethod
    def read(stream: core.Data, gv: core.GameVersion) -> LoginBonus:
        if gv < 80000:
            logins_old = LoginSets.read(stream)
            return LoginBonus(logins_old)
        else:
            total = stream.read_int()
            logins: dict[int, Login] = {}
            for _ in range(total):
                id = stream.read_int()
                logins[id] = Login.read(stream)
            return LoginBonus(logins=logins)

    def write(self, stream: core.Data, gv: core.GameVersion):
        if gv < 80000:
            (self.old_logins or LoginSets([])).write(stream)
        elif gv >= 80000:
            logins = self.logins or {}
            stream.write_int(len(logins))
            for id, login in logins.items():
                stream.write_int(id)
                login.write(stream)

    def serialize(
        self,
    ) -> dict[str, Any]:
        if self.old_logins is not None:
            return {"old_logins": self.old_logins.serialize()}
        elif self.logins is not None:
            return {
                "logins": {
                    id: login.serialize() for id, login in self.logins.items()
                }
            }
        else:
            return {}

    @staticmethod
    def deserialize(data: dict[str, Any]) -> LoginBonus:
        if "old_logins" in data:
            return LoginBonus(
                old_logins=LoginSets.deserialize(data["old_logins"])
            )
        elif "logins" in data:
            return LoginBonus(
                logins={
                    int(id): Login.deserialize(login)
                    for id, login in data["logins"].items()
                }
            )
        else:
            return LoginBonus()

    def __repr__(self):
        return f"LoginBonus({self.old_logins}, {self.logins})"

    def __str__(self):
        return f"LoginBonus({self.old_logins}, {self.logins})"

    def get_login(self, id: int) -> Login | None:
        if self.logins is not None:
            return self.logins.get(id)
        else:
            return None



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import dialog_creator


class Stage:
    def __init__(self, score: int):
        self.score = score

    @staticmethod
    def init() -> Stage:
        return Stage(0)

    @staticmethod
    def read(stream: core.Data) -> Stage:
        score = stream.read_int()
        return Stage(score)

    def write(self, stream: core.Data):
        stream.write_int(self.score)

    def serialize(self) -> int:
        return self.score

    @staticmethod
    def deserialize(data: int) -> Stage:
        return Stage(data)

    def __repr__(self) -> str:
        return f"Stage(score={self.score!r})"

    def __str__(self) -> str:
        return f"Stage(score={self.score!r})"


class Chapter:
    def __init__(self, stages: dict[int, Stage]):
        self.stages = stages

    def get_stage(self, stage_id: int) -> Stage:
        if stage_id not in self.stages:
            self.stages[stage_id] = Stage.init()
        return self.stages[stage_id]

    @staticmethod
    def init() -> Chapter:
        return Chapter({})

    @staticmethod
    def read(stream: core.Data) -> Chapter:
        total = stream.read_int()
        stages: dict[int, Stage] = {}
        for _ in range(total):
            stage_id = stream.read_int()
            stage = Stage.read(stream)
            stages[stage_id] = stage

        return Chapter(stages)

    def write(self, stream: core.Data):
        stream.write_int(len(self.stages))
        for stage_id, stage in self.stages.items():
            stream.write_int(stage_id)
            stage.write(stream)

    def serialize(self) -> dict[int, Any]:
        return {stage_id: stage.serialize() for stage_id, stage in self.stages.items()}

    @staticmethod
    def deserialize(data: dict[int, Any]) -> Chapter:
        return Chapter(
            {stage_id: Stage.deserialize(stage) for stage_id, stage in data.items()}
        )

    def __repr__(self) -> str:
        return f"Chapter(stages={self.stages!r})"

    def __str__(self) -> str:
        return f"Chapter(stages={self.stages!r})"


class Chapters:
    def __init__(self, chapters: dict[int, Chapter]):
        self.chapters = chapters

    def get_stage(self, chapter_id: int, stage_id: int) -> Stage:
        if chapter_id not in self.chapters:
            self.chapters[chapter_id] = Chapter.init()
        return self.chapters[chapter_id].get_stage(stage_id)

    @staticmethod
    def init() -> Chapters:
        return Chapters({})

    @staticmethod
    def read(stream: core.Data) -> Chapters:
        total = stream.read_int()
        chapters: dict[int, Chapter] = {}
        for _ in range(total):
            chapter_id = stream.read_int()
            chapter = Chapter.read(stream)
            chapters[chapter_id] = chapter

        return Chapters(chapters)

    def write(self, stream: core.Data):
        stream.write_int(len(self.chapters))
        for chapter_id, chapter in self.chapters.items():
            stream.write_int(chapter_id)
            chapter.write(stream)

    def serialize(self) -> dict[int, Any]:
        return {
            chapter_id: chapter.serialize()
            for chapter_id, chapter in self.chapters.items()
        }

    @staticmethod
    def deserialize(data: dict[int, Any]) -> Chapters:
        return Chapters(
            {
                chapter_id: Chapter.deserialize(chapter)
                for chapter_id, chapter in data.items()
            }
        )

    def __repr__(self) -> str:
        return f"Chapters(chapters={self.chapters!r})"

    def __str__(self) -> str:
        return f"Chapters(chapters={self.chapters!r})"


class Ranking:
    def __init__(
        self,
        score: int,
        ranking: int,
        has_submitted: bool,
        has_completed: bool,
        has_seen_results: bool,
        start_date: int,
        end_date: int,
        event_number: int,
        should_show_rank_description: bool,
        should_show_start_message: bool,
        submit_error_flag: bool,
        other: str | None,
    ):
        self.score = score
        self.ranking = ranking
        self.has_submitted = has_submitted
        self.has_completed = has_completed
        self.has_seen_results = has_seen_results
        self.start_date = start_date
        self.end_date = end_date
        self.event_number = event_number
        self.should_show_rank_description = should_show_rank_description
        self.should_show_start_message = should_show_start_message
        self.submit_error_flag = submit_error_flag
        self.did_win_rewards = False
        self.other = other

    @staticmethod
    def init() -> Ranking:
        return Ranking(
            0,
            0,
            False,
            False,
            False,
            0,
            0,
            0,
            False,
            False,
            False,
            None,
        )

    @staticmethod
    def read(stream: core.Data, game_version: core.GameVersion) -> Ranking:
        score = stream.read_int()
        ranking = stream.read_int()
        has_submitted = stream.read_bool()
        has_completed = stream.read_bool()
        has_seen_results = stream.read_bool()
        start_date = stream.read_int()
        end_date = stream.read_int()
        event_number = stream.read_int()
        should_show_rank_description = stream.read_bool()
        should_show_start_message = stream.read_bool()
        submit_error_flag = stream.read_bool()

        if game_version >= 140500:
            # game seems to do more that just this, may break in the future
            other = stream.read_string()
        else:
            other = None
        return Ranking(
            score,
            ranking,
            has_submitted,
            has_completed,
            has_seen_results,
            start_date,
            end_date,
            event_number,
            should_show_rank_description,
            should_show_start_message,
            submit_error_flag,
            other,
        )

    def write(self, stream: core.Data, game_version: core.GameVersion):
        stream.write_int(self.score)
        stream.write_int(self.ranking)
        stream.write_bool(self.has_submitted)
        stream.write_bool(self.has_completed)
        stream.write_bool(self.has_seen_results)
        stream.write_int(self.start_date)
        stream.write_int(self.end_date)
        stream.write_int(self.event_number)
        stream.write_bool(self.should_show_rank_description)
        stream.write_bool(self.should_show_start_message)
        stream.write_bool(self.submit_error_flag)
        if game_version >= 140500:
            # game seems to do more that just this, may break in the future
            stream.write_string(self.other or "")

    def read_did_win_rewards(self, stream: core.Data):
        self.did_win_rewards = stream.read_bool()

    def write_did_win_rewards(self, stream: core.Data):
        stream.write_bool(self.did_win_rewards)

    def serialize(self) -> dict[str, Any]:
        return {
            "score": self.score,
            "ranking": self.ranking,
            "has_submitted": self.has_submitted,
            "has_completed": self.has_completed,
            "has_seen_results": self.has_seen_results,
            "start_date": self.start_date,
            "end_date": self.end_date,
            "event_number": self.event_number,
            "should_show_rank_description": self.should_show_rank_description,
            "should_show_start_message": self.should_show_start_message,
            "submit_error_flag": self.submit_error_flag,
            "did_win_rewards": self.did_win_rewards,
            "other": self.other,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Ranking:
        ranking = Ranking(
            data.get("score", 0),
            data.get("ranking", 0),
            data.get("has_submitted", False),
            data.get("has_completed", False),
            data.get("has_seen_results", False),
            data.get("start_date", 0),
            data.get("end_date", 0),
            data.get("event_number", 0),
            data.get("should_show_rank_description", False),
            data.get("should_show_start_message", False),
            data.get("submit_error_flag", False),
            data.get("other", None),
        )
        ranking.did_win_rewards = data.get("did_win_rewards", False)
        return ranking

    def __repr__(self) -> str:
        return (
            f"Ranking(score={self.score!r}, ranking={self.ranking!r}, "
            f"has_submitted={self.has_submitted!r}, has_completed={self.has_completed!r}, "
            f"has_seen_results={self.has_seen_results!r}, start_date={self.start_date!r}, "
            f"end_date={self.end_date!r}, event_number={self.event_number!r}, "
            f"should_show_rank_description={self.should_show_rank_description!r}, "
            f"should_show_start_message={self.should_show_start_message!r}, "
            f"submit_error_flag={self.submit_error_flag!r},"
            f"did_win_rewards={self.did_win_rewards!r}),"
            f"other={self.other!r})"
        )

    def __str__(self) -> str:
        return self.__repr__()


class Dojo:
    def __init__(self, chapters: Chapters):
        self.chapters = chapters
        self.item_lock_flags = False
        self.item_locks = [False] * 6
        self.ranking = Ranking.init()

    @staticmethod
    def init() -> Dojo:
        return Dojo(Chapters.init())

    @staticmethod
    def read_chapters(stream: core.Data) -> Dojo:
        chapters = Chapters.read(stream)
        return Dojo(chapters)

    def write_chapters(self, stream: core.Data):
        self.chapters.write(stream)

    def read_item_locks(self, stream: core.Data):
        self.item_lock_flags = stream.read_bool()
        self.item_locks = stream.read_bool_list(6)

    def write_item_locks(self, stream: core.Data):
        stream.write_bool(self.item_lock_flags)
        stream.write_bool_list(self.item_locks, write_length=False, length=6)

    def read_ranking(self, stream: core.Data, game_version: core.GameVersion):
        self.ranking = Ranking.read(stream, game_version)

    def write_ranking(self, stream: core.Data, game_version: core.GameVersion):
        self.ranking.write(stream, game_version)

    def serialize(self) -> dict[str, Any]:
        return {
            "chapters": self.chapters.serialize(),
            "item_locks": self.item_locks,
            "item_lock_flags": self.item_lock_flags,
            "ranking": self.ranking.serialize(),
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Dojo:
        chapters = Chapters.deserialize(data.get("chapters", {}))
        item_locks = data.get("item_locks", [])
        item_lock_flags = data.get("item_lock_flags", False)
        dojo = Dojo(chapters)
        dojo.item_locks = item_locks
        dojo.item_lock_flags = item_lock_flags
        dojo.ranking = Ranking.deserialize(data.get("ranking", {}))
        return dojo

    def __repr__(self) -> str:
        return f"Dojo(chapters={self.chapters!r}, item_locks={self.item_locks!r}, item_lock_flags={self.item_lock_flags!r}, ranking={self.ranking!r})"

    def __str__(self) -> str:
        return self.__repr__()

    def edit_score(self):
        stage = self.chapters.get_stage(0, 0)
        stage.score = dialog_creator.SingleEditor(
            "dojo_score",
            stage.score,
            None,
            localized_item=True,
        ).edit()


def edit_dojo_score(save_file: core.SaveFile):
    save_file.dojo.edit_score()



from __future__ import annotations
from typing import Any
from bcsfe import core


class MySale:
    def __init__(self, dict_1: dict[int, int], dict_2: dict[int, bool]):
        self.dict_1 = dict_1
        self.dict_2 = dict_2

    @staticmethod
    def init() -> MySale:
        return MySale({}, {})

    @staticmethod
    def read_bonus_hash(stream: core.Data):
        variable_length = stream.read_variable_length_int()
        dict_1 = {}
        for _ in range(variable_length):
            key = stream.read_variable_length_int()
            value = stream.read_variable_length_int()
            dict_1[key] = value

        variable_length = stream.read_variable_length_int()
        dict_2 = {}
        for _ in range(variable_length):
            key = stream.read_variable_length_int()
            value = stream.read_byte()
            dict_2[key] = value

        return MySale(dict_1, dict_2)

    def write_bonus_hash(self, stream: core.Data):
        stream.write_variable_length_int(len(self.dict_1))
        for key, value in self.dict_1.items():
            stream.write_variable_length_int(key)
            stream.write_variable_length_int(value)

        stream.write_variable_length_int(len(self.dict_2))
        for key, value in self.dict_2.items():
            stream.write_variable_length_int(key)
            stream.write_byte(value)

    def serialize(self) -> dict[str, Any]:
        return {
            "dict_1": self.dict_1,
            "dict_2": self.dict_2,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> MySale:
        return MySale(data.get("dict_1", {}), data.get("dict_2", {}))

    def __repr__(self) -> str:
        return f"MySale(dict_1={self.dict_1}, dict_2={self.dict_2})"

    def __str__(self) -> str:
        return f"MySale(dict_1={self.dict_1}, dict_2={self.dict_2})"



from __future__ import annotations
from typing import Any
from bcsfe import core
from bcsfe.cli import edits, color


class Stage:
    def __init__(self, clear_times: int):
        self.clear_times = clear_times

    @staticmethod
    def init() -> Stage:
        return Stage(0)

    @staticmethod
    def read(data: core.Data) -> Stage:
        clear_times = data.read_short()
        return Stage(clear_times)

    def write(self, data: core.Data):
        data.write_short(self.clear_times)

    def serialize(self) -> int:
        return self.clear_times

    @staticmethod
    def deserialize(data: int) -> Stage:
        return Stage(
            data,
        )

    def __repr__(self):
        return f"Stage({self.clear_times})"

    def __str__(self):
        return self.__repr__()

    def clear_stage(self, clear_amount: int = 1, ensure_cleared_only: bool = False):
        if ensure_cleared_only:
            self.clear_times = self.clear_times or clear_amount
        else:
            self.clear_times = clear_amount

    def unclear_stage(self):
        self.clear_times = 0


class Chapter:
    def __init__(
        self,
        selected_stage: int,
        clear_progress: int,
        unlock_state: int,
        stages: list[Stage],
    ):
        self.selected_stage = selected_stage
        self.clear_progress = clear_progress
        self.unlock_state = unlock_state
        self.stages = stages

        self.total_stages = 0

    def clear_stage(
        self,
        index: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
        ensure_cleared_only: bool = False,
    ) -> bool:
        if overwrite_clear_progress:
            self.clear_progress = index + 1
        else:
            self.clear_progress = max(self.clear_progress, index + 1)
        self.stages[index].clear_stage(clear_amount, ensure_cleared_only)
        self.chapter_unlock_state = 3
        if index == self.total_stages - 1:
            return True
        return False

    def unclear_stage(self, index: int) -> bool:
        self.clear_progress = min(self.clear_progress, index)
        self.stages[index].unclear_stage()
        return True

    @staticmethod
    def init() -> Chapter:
        return Chapter(0, 0, 0, [])

    @staticmethod
    def read(data: core.Data) -> Chapter:
        selected_stage = data.read_byte()
        clear_progress = data.read_byte()
        unlock_state = data.read_byte()
        total_stages = data.read_short()
        stages = [Stage.read(data) for _ in range(total_stages)]
        return Chapter(
            selected_stage,
            clear_progress,
            unlock_state,
            stages,
        )

    def write(self, data: core.Data):
        data.write_byte(self.selected_stage)
        data.write_byte(self.clear_progress)
        data.write_byte(self.unlock_state)
        data.write_short(len(self.stages))
        for stage in self.stages:
            stage.write(data)

    def serialize(self) -> dict[str, Any]:
        return {
            "selected_stage": self.selected_stage,
            "clear_progress": self.clear_progress,
            "unlock_state": self.unlock_state,
            "stages": [stage.serialize() for stage in self.stages],
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Chapter:
        return Chapter(
            data.get("selected_stage", 0),
            data.get("clear_progress", 0),
            data.get("unlock_state", 0),
            [Stage.deserialize(stage) for stage in data.get("stages", [])],
        )

    def __repr__(self):
        return f"Chapter({self.selected_stage}, {self.clear_progress}, {self.unlock_state}, {self.stages})"

    def __str__(self):
        return self.__repr__()


class ChaptersStars:
    def __init__(self, unknown: int, chapters: list[Chapter]):
        self.unknown = unknown
        self.chapters = chapters

    def clear_stage(
        self,
        star: int,
        stage: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
        ensure_cleared_only: bool = False,
    ) -> bool:
        finished = self.chapters[star].clear_stage(
            stage, clear_amount, overwrite_clear_progress, ensure_cleared_only
        )
        if finished:
            if star + 1 < len(self.chapters):
                self.chapters[star + 1].chapter_unlock_state = 1
        return finished

    def unclear_stage(self, star: int, stage: int) -> bool:
        finished = self.chapters[star].unclear_stage(stage)
        if finished and star + 1 < len(self.chapters):
            for chapter in self.chapters[star + 1 :]:
                chapter.chapter_unlock_state = 0
        return finished

    @staticmethod
    def init() -> ChaptersStars:
        return ChaptersStars(0, [])

    @staticmethod
    def read(data: core.Data) -> ChaptersStars:
        unknown = data.read_byte()
        total_stars = data.read_byte()
        chapters = [Chapter.read(data) for _ in range(total_stars)]
        return ChaptersStars(
            unknown,
            chapters,
        )

    def write(self, data: core.Data):
        data.write_byte(self.unknown)
        data.write_byte(len(self.chapters))
        for chapter in self.chapters:
            chapter.write(data)

    def serialize(self) -> dict[str, Any]:
        return {
            "unknown": self.unknown,
            "chapters": [chapter.serialize() for chapter in self.chapters],
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> ChaptersStars:
        return ChaptersStars(
            data.get("unknown", 0),
            [Chapter.deserialize(chapter) for chapter in data.get("chapters", [])],
        )

    def __repr__(self):
        return f"ChaptersStars({self.unknown}, {self.chapters})"

    def __str__(self):
        return self.__repr__()


class ZeroLegendsChapters:
    def __init__(self, chapters: list[ChaptersStars]):
        self.chapters = chapters

    def clear_stage(
        self,
        map: int,
        star: int,
        stage: int,
        clear_amount: int = 1,
        overwrite_clear_progress: bool = False,
        ensure_cleared_only: bool = False,
    ) -> bool:
        self.create(map)
        finished = self.chapters[map].clear_stage(
            star, stage, clear_amount, overwrite_clear_progress, ensure_cleared_only
        )
        if finished and map + 1 < len(self.chapters):
            self.chapters[map + 1].chapters[0].chapter_unlock_state = 1

        return finished

    def unclear_stage(self, map: int, star: int, stage: int) -> bool:
        self.create(map)
        finished = self.chapters[map].unclear_stage(star, stage)
        if finished and map + 1 < len(self.chapters) and star == 0:
            for chapter in self.chapters[map + 1].chapters:
                chapter.chapter_unlock_state = 0

        return finished

    @staticmethod
    def init() -> ZeroLegendsChapters:
        return ZeroLegendsChapters([])

    @staticmethod
    def read(data: core.Data) -> ZeroLegendsChapters:
        total_chapters = data.read_short()
        chapters = [ChaptersStars.read(data) for _ in range(total_chapters)]
        return ZeroLegendsChapters(
            chapters,
        )

    def write(self, data: core.Data):
        data.write_short(len(self.chapters))
        for chapter in self.chapters:
            chapter.write(data)

    def serialize(self) -> list[dict[str, Any]]:
        return [chapter.serialize() for chapter in self.chapters]

    @staticmethod
    def deserialize(data: list[dict[str, Any]]) -> ZeroLegendsChapters:
        return ZeroLegendsChapters(
            [ChaptersStars.deserialize(chapter) for chapter in data],
        )

    def __repr__(self):
        return f"Chapters({self.chapters})"

    def __str__(self):
        return self.__repr__()

    def get_total_stars(self, chapter_id: int) -> int:
        return len(self.chapters[chapter_id].chapters)

    def get_total_stages(self, chapter_id: int, star: int) -> int:
        return len(self.chapters[chapter_id].chapters[star].stages)

    def create(self, chapter_id: int):
        diff = chapter_id - len(self.chapters)

        if diff >= 0:
            for _ in range(diff + 1):
                stages = [Stage(0)] * self.get_total_stages(0, 0)
                chapters = [Chapter(0, 0, 0, stages)] * self.get_total_stars(0)
                chapters_stars = ChaptersStars(0, chapters)
                self.chapters.append(chapters_stars)

    @staticmethod
    def edit_zero_legends(save_file: core.SaveFile):
        color.ColoredText.localize("zero_legends_warning")
        zero_legends_chapters = save_file.zero_legends
        zero_legends_chapters.edit_chapters(save_file, "ND", base_index=34000)

    @staticmethod
    def edit_catclaw_championships(save_file: core.SaveFile):
        zero_legends_chapters = save_file.dojo_chapters
        zero_legends_chapters.edit_chapters(save_file, "G", 37000, True)

    def edit_chapters(
        self,
        save_file: core.SaveFile,
        letter_code: str,
        base_index: int,
        no_r_prefix: bool = False,
    ):
        edits.map.edit_chapters(
            save_file, self, letter_code, no_r_prefix=no_r_prefix, base_index=base_index
        )

    def unclear_rest(self, stages: list[int], stars: int, id: int):
        if not stages:
            return
        for star in range(stars, self.get_total_stars(id)):
            for stage in range(max(stages), self.get_total_stages(id, star)):
                self.chapters[id].chapters[star].stages[stage].clear_times = 0
                self.chapters[id].chapters[star].clear_progress = 0

    def set_total_stages(self, map: int, total_stages: int):
        self.create(map)
        for chapter in self.chapters[map].chapters:
            chapter.total_stages = total_stages
